# -*- coding: utf-8 -*-
"""
/*------------------------------------------------------*
|                         BRAILS                        |
|                                                       |
| Author: Charles Wang,  UC Berkeley, c_w@berkeley.edu  |
|                                                       |
| Date:    01/5/2021                                    |
*------------------------------------------------------*/
"""

import argparse
import os
from pathlib import Path
import random
import numpy as np
from glob import glob
import pandas as pd
import geopandas as gpd
import json


from brails.modules import RoofClassifier, OccupancyClassifier, SoftstoryClassifier, FoundationHeightClassifier, YearBuiltClassifier
from brails.modules import FoundationHeightClassifier as NFloorDetector # pseudo as a placeholder, will change when available

from .workflow.Footprints import getMSFootprintsByPlace, getStateNameByBBox, getOSMFootprints, getMSFootprintsByBbox
from .workflow.Images import getGoogleImages



class CityBuilder:
    """Class for creating city-scale BIM."""

    def __init__(self, attributes=['story','occupancy','roofshape'], numBldg=10, random=True,bbox=[], place='', footPrints='OSM', save=True, fileName='', workDir='tmp',GoogleMapAPIKey='', overwrite=False, reDownloadImgs=False):
        """init function for CityBuilder class.

        Args:

            attributes (list): A list of building attributes, such as ['story', 'occupancy', 'roofshape'], which are available in the current version.
            numBldg (int): Number of buildings to generate.
            random (bool): Randomly select numBldg buildings from the database if random is True
            bbox (list): [north, west, south, east]
            place (str): The region of interest, e.g., Berkeley, California
            footPrints (str): The footprint provide, choose from OSM or Microsoft. The default value is OSM.
            save (bool): Save temporary files. Default value is True.
            fileName (str): Name of the generated BIM file. Default value will be generated if not provided by the user.
            workDir (str): Work directory where all files will be saved. Default value is ./tmp
            GoogleMapAPIKey (str): Google API Key
            overwrite (bool): Overwrite existing tmp files. Default value is False.
            reDownloadImgs (bool): Re-download if an image exists locally. Default value is False.

        Returns:
            BIM (geopandas dataframe): BIM
            fileName (str): Path to the BIM file.

        """

        if GoogleMapAPIKey == '':
            print('Please provide GoogleMapAPIKey') 
            return None

        if not os.path.exists(workDir): os.makedirs(workDir)

        #if fileName=='': fileName=Path(f'{place} {state}_BIM.geojson'.replace(' ','_').replace(',','_'))
        if fileName=='': 

            if len(bbox)>0: 
                fileName = Path(f'/{bbox[0]}_{bbox[1]}_{bbox[2]}_{bbox[3]}_BIM.geojson')
            else:
                fileName = Path(f'/{place}_BIM.geojson'.replace(' ','_').replace(',','_'))

        

        bimFile = Path(f'{workDir}/{fileName}')
        
        self.attributes = attributes
        self.numBldg = numBldg
        self.random = random
        self.place = place
        self.save = save
        self.fileName = fileName
        self.workDir = workDir
        self.overwrite =  overwrite
        self.bimFile = bimFile
        self.GoogleMapAPIKey = GoogleMapAPIKey
        self.reDownloadImgs = reDownloadImgs
        self.footPrints = footPrints

        if os.path.exists(bimFile) and not overwrite:
            print('{} already exists.'.format(bimFile))
            BIM = gpd.read_file(bimFile)
            return None



        if len(bbox)>0: # bbox provided
            self.getFootprintByBbox(bbox=bbox,workDir=workDir,GoogleMapAPIKey=GoogleMapAPIKey,save=save,overwrite=overwrite)

        elif len(place)>0:
            print(place)
            self.getFootprintByPlace(place=place,workDir=workDir,GoogleMapAPIKey=GoogleMapAPIKey,save=save,overwrite=overwrite)

        else:
            assert False, f"Please provide a bbox or place name."
        
        print(f'{self.BIM.shape[0]} buildings found.')
        if numBldg < self.BIM.shape[0]: 
            if random: self.BIM = self.BIM.sample(n=numBldg)
            else: self.BIM = self.BIM[:numBldg]
            print(f'{self.BIM.shape[0]} buildings selected.')
        
    def getFootprintByBbox(self,bbox=[],workDir='',GoogleMapAPIKey='',save=True,overwrite=True):
        #BIM = getCityBIM()

        stateName = getStateNameByBBox(bbox=bbox,workDir=workDir,GoogleMapAPIKey=GoogleMapAPIKey) # isNone means didn't find in the U.S.
        if stateName:# bbox, in the US
            #self.BIM, footprintFilename = getMSFootprintsByBbox(bbox=bbox,stateName=stateName,save=save,workDir=workDir,overwrite=overwrite,GoogleMapAPIKey=GoogleMapAPIKey)
            if self.footPrints=='OSM':
                print(f"Trying to get OSM footprints by the bbox {bbox}.")
                self.BIM, footprintFilename = getOSMFootprints(bbox=bbox,save=save,workDir=workDir,overwrite=overwrite)
            elif self.footPrints=='Microsoft':
                # MS
                print(f"Trying to get MS footprints by the bbox {bbox}.")
                self.BIM, footprintFilename = getMSFootprintsByBbox(bbox=bbox,stateName=stateName,save=save,workDir=workDir,overwrite=overwrite,GoogleMapAPIKey=GoogleMapAPIKey)
            else:
                assert False, f"footPrints must be OSM or Microsoft in CityBuilder. Your value is {self.footPrints}."
        else: # out of US, use OSM no matter what
                print(f"The bbox seems to be outside of the US. Footprints probably only exist in OSM.  Trying to get OSM footprints by the bbox {bbox}.")
                self.BIM, footprintFilename = getOSMFootprints(bbox=bbox,save=save,workDir=workDir,overwrite=overwrite)
            

    def getFootprintByPlace(self,place='',workDir='',GoogleMapAPIKey='',save=True,overwrite=True):
        if self.footPrints=='OSM':
            print(f"Trying to get OSM footprints by the place name {place}.")
            self.BIM, footprintFilename = getOSMFootprints(place=place,save=save,workDir=workDir,overwrite=True)
        elif self.footPrints=='Microsoft':
            #assert False, f"Trying to get Microsoft footprints by the place name {place}."
            self.BIM, footprintFilename = getMSFootprintsByPlace(place=place,save=save,workDir=workDir,GoogleMapAPIKey=GoogleMapAPIKey,overwrite=overwrite)
        else: 
            assert False, f"footPrints must be OSM or Microsoft in CityBuilder. Your value is {self.footPrints}."

    def downloadImgs(self,fov=60,pitch=0): 

        if os.path.exists(self.bimFile) and not self.overwrite:
            print('{} already exists.'.format(self.bimFile))
            BIM = gpd.read_file(self.bimFile)
            return BIM 
        
        if self.BIM.shape[0] < 1:
            print('Can not build because the Footprints module didn\'t find footprints for this region.')
            return None

        print('Starting downloading images.')
        imgDir = os.path.join(self.workDir, 'images')

        #imageTypes = ['StreetView','TopView']
        imageTypes = []
        if 'roofshape' in self.attributes:
            imageTypes.append('TopView')

        svAttrs = ['occupancy','softstory','elevated','story','year']
        usvAttrs = [x for x in svAttrs if x in set(self.attributes)]
        if len(usvAttrs)> 0: imageTypes.append('StreetView')

        self.imageTypes = imageTypes
        self.imgDir = imgDir

        getGoogleImages(self.BIM,GoogleMapAPIKey=self.GoogleMapAPIKey, imageTypes=imageTypes, imgDir=imgDir, ncpu=2,fov=fov,pitch=pitch,reDownloadImgs=self.reDownloadImgs)


    def build(self): 

        self.downloadImgs()

        imageTypes = self.imageTypes
        imgDir = self.imgDir

        #BIM = allFootprints
        self.BIM['Lon'] = self.BIM['geometry'].centroid.x#.round(decimals=6)
        self.BIM['Lat'] = self.BIM['geometry'].centroid.y#.round(decimals=6)
        for cat in imageTypes:
            self.BIM[cat] = self.BIM.apply(lambda row: Path(f"{imgDir}/{cat}/{cat}x{'%.6f'%row['Lon']}x{'%.6f'%row['Lat']}.png"), axis=1)

        self.BIM.reset_index(inplace = True) 
        self.BIM['ID'] = self.BIM.index
    

        for attr in self.attributes:
            if attr.lower()=='roofshape':

                # initialize a roof classifier
                roofModel = RoofClassifier(printRes=False)

                # use the roof classifier 
                roofShape_df = roofModel.predict(self.BIM['TopView'].tolist())

                roofShape = roofShape_df['prediction'].to_list()
                roofShapeProb = roofShape_df['probability'].to_list()
                self.BIM['roofShape'] = self.BIM.apply(lambda x: roofShape[x['ID']], axis=1)
                self.BIM['roofShapeProb'] = self.BIM.apply(lambda x: roofShapeProb[x['ID']], axis=1)

            elif attr.lower()=='occupancy':
                # initialize an occupancy classifier
                occupancyModel = OccupancyClassifier(printRes=False)
                # use the occupancy classifier 
                occupancy_df = occupancyModel.predict(self.BIM['StreetView'].tolist())

                occupancy = occupancy_df['prediction'].to_list()
                occupancyProb = occupancy_df['probability'].to_list()
                self.BIM['occupancy'] = self.BIM.apply(lambda x: occupancy[x['ID']], axis=1)
                self.BIM['occupancyProb'] = self.BIM.apply(lambda x: occupancyProb[x['ID']], axis=1)

            elif attr.lower()=='softstory':
                # initialize a soft-story classifier
                ssModel = SoftstoryClassifier(printRes=False)
                # use the softstory classifier 
                softstory_df = ssModel.predict(self.BIM['StreetView'].tolist())

                softstory = softstory_df['prediction'].to_list()
                softstoryProb = softstory_df['probability'].to_list()
                self.BIM['softStory'] = self.BIM.apply(lambda x: softstory[x['ID']], axis=1)
                self.BIM['softStoryProb'] = self.BIM.apply(lambda x: softstoryProb[x['ID']], axis=1)

            elif attr.lower()=='elevated':
                # initialize a foundation classifier
                elvModel = FoundationHeightClassifier(workDir=self.workDir, maskBuildings=False)

                # use the classifier 
                elv_df = elvModel.predict(self.BIM['StreetView'].tolist())

                elv = elv_df['prediction'].to_list()
                elvProb = elv_df['probability'].to_list()
                self.BIM['elevated'] = self.BIM.apply(lambda x: elv[x['ID']], axis=1)
                self.BIM['elevatedProb'] = self.BIM.apply(lambda x: elvProb[x['ID']], axis=1) 

            elif attr.lower()=='story':
                storyModel = NFloorDetector()

                story_df = storyModel.predict(self.BIM['StreetView'].tolist())

                story = story_df['prediction'].to_list()
                storyProb = story_df['probability'].to_list()
                self.BIM['story'] = self.BIM.apply(lambda x: story[x['ID']], axis=1)
                self.BIM['storyProb'] = self.BIM.apply(lambda x: storyProb[x['ID']], axis=1) 

            elif attr.lower()=='year':
                yearModel = YearBuiltClassifier(workDir=self.workDir)

                year_df = yearModel.predict(self.BIM['StreetView'].tolist())

                year = year_df['prediction'].to_list()
                yearProb = year_df['probability'].to_list()
                self.BIM['year'] = self.BIM.apply(lambda x: year[x['ID']], axis=1)
                self.BIM['yearProb'] = self.BIM.apply(lambda x: yearProb[x['ID']], axis=1) 

            elif attr.lower()=='softstory2':
                from brails.modules.SoftstoryDetector.SoftstoryDetector import SoftstoryDetector
                # initialize a soft-story detector
                ssModel = SoftstoryDetector()
                # use the soft-story detector 
                softstory_df = ssModel.detect(self.BIM['StreetView'].tolist())

                softstory = softstory_df['prediction'].to_list()
                softstoryProb = softstory_df['probability'].to_list()
                self.BIM['softStory2'] = self.BIM.apply(lambda x: softstory[x['ID']], axis=1)
                self.BIM['softStory2Prob'] = self.BIM.apply(lambda x: softstoryProb[x['ID']], axis=1)

                self.BIM['StreetView2'] = self.BIM['StreetView'].astype(str) 
                

            else:
                assert False, f"attributes can only contain occupancy, roofshape, softstory. Your {attr} caused an error."

        # delete columns
        self.BIM.drop(columns=['Lat','Lon','index'], axis=1, inplace=True)
        for c in imageTypes:
            self.BIM.drop(columns=[c], axis=1, inplace=True)

        # save
        self.BIM.to_file(self.bimFile, driver='GeoJSON')
        print('BIM saved at {}'.format(self.bimFile))

        #print(self.BIM)
        return self.BIM








def main():
    pass

if __name__ == '__main__':
    main()







from __future__ import print_function
from PIL import Image
import torchvision.datasets as datasets
import torch.utils.data as data

class MNISTInstance(datasets.MNIST):
    """MNIST Instance Dataset.
    """

    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        if self.train:
            img, target = self.train_data[index], self.train_labels[index]
        else:
            img, target = self.test_data[index], self.test_labels[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img.numpy(), mode='L')

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target, index



import torch
from torch.autograd import Function
from torch import nn
from .alias_multinomial import AliasMethod
import math
import socket

if socket.gethostname() == 'ncc1701e':
    device = 'cpu'
else:
    device = 'cuda' if torch.cuda.is_available() else 'cpu'

class NCEFunction(Function):
    @staticmethod
    def forward(self, x, y, memory, idx, params):
        K = int(params[0].item())
        T = params[1].item()
        Z = params[2].item()

        momentum = params[3].item()
        batchSize = x.size(0)
        outputSize = memory.size(0)
        inputSize = memory.size(1)

        # sample positives & negatives
        idx.select(1,0).copy_(y.data)

        # sample correspoinding weights
        weight = torch.index_select(memory, 0, idx.view(-1))
        weight.resize_(batchSize, K+1, inputSize)

        # inner product
        out = torch.bmm(weight, torch.reshape(x, (batchSize, inputSize, 1)))
        #out = torch.bmm(weight, x.data.resize_(batchSize, inputSize, 1)) # Bugfix
        out.div_(T).exp_() # batchSize * self.K+1
        x.data.resize_(batchSize, inputSize)

        if Z < 0:
            params[2] = out.mean() * outputSize
            Z = params[2].item()
            print("normalization constant Z is set to {:.1f}".format(Z))

        out.div_(Z).resize_(batchSize, K+1)

        self.save_for_backward(x, memory, y, weight, out, params)

        return out

    @staticmethod
    def backward(self, gradOutput):
        x, memory, y, weight, out, params = self.saved_tensors
        K = int(params[0].item())
        T = params[1].item()
        Z = params[2].item()
        momentum = params[3].item()
        batchSize = gradOutput.size(0)
        
        # gradients d Pm / d linear = exp(linear) / Z
        gradOutput.data.mul_(out.data)
        # add temperature
        gradOutput.data.div_(T)

        gradOutput = torch.reshape(gradOutput,(batchSize, 1, K + 1))
        #gradOutput.data.resize_(batchSize, 1, K+1) # Another bugfix
        
        # gradient of linear
        gradInput = torch.bmm(gradOutput.data, weight)
        gradInput.resize_as_(x)

        # update the non-parametric data
        weight_pos = weight.select(1, 0).resize_as_(x)
        weight_pos.mul_(momentum)
        weight_pos.add_(torch.mul(x.data, 1 - momentum))
        w_norm = weight_pos.pow(2).sum(1, keepdim=True).pow(0.5)
        updated_weight = weight_pos.div(w_norm)
        memory.index_copy_(0, y, updated_weight)
        
        return gradInput, None, None, None, None
''''''
class NCEAverage(nn.Module):

    def __init__(self, inputSize, outputSize, K, T=0.07, momentum=0.5, Z=None):
        super(NCEAverage, self).__init__()
        self.nLem = outputSize
        self.unigrams = torch.ones(self.nLem)
        self.multinomial = AliasMethod(self.unigrams)
        self.multinomial.cuda()
        self.K = K

        self.register_buffer('params',torch.tensor([K, T, -1, momentum]));
        stdv = 1. / math.sqrt(inputSize/3)
        self.register_buffer('memory', torch.rand(outputSize, inputSize).mul_(2*stdv).add_(-stdv))
 
    def forward(self, x, y):
        batchSize = x.size(0)
        idx = self.multinomial.draw(batchSize * (self.K+1)).view(batchSize, -1)
        out = NCEFunction.apply(x, y, self.memory, idx, self.params)
        return out




from models.attention_module import *
from models.basic_layer import ResidualBlock, ResidualBlock1D

class DenseBlock(nn.Module):
    def __init__(self, input_size, output_size, last_layer=False, dropout_p=0.0):
        super(DenseBlock, self).__init__()
        self.fc = nn.Linear(input_size, output_size)
        self.dp = nn.Dropout(dropout_p)
        self.bn = nn.BatchNorm1d(output_size)
        self.relu = nn.ReLU()
        self.last_layer = last_layer

        # initialize weights
        # x = float(1/128)
        # torch.nn.init.uniform_(self.fc.weight,-x,x)
        # torch.nn.init.uniform_(self.fc.bias,-x,x)



    def forward(self, x):
        out = self.fc(x)
        out = self.dp(out)
        if not self.last_layer:
            out = self.bn(out)
            out = self.relu(out)
        return out


class MLP(nn.Module):
    # Lesson learned: Dropout only on the last layer seems to be superior
    # hidden size 1024 or at least double that is best
    def __init__(self, input_size, hidden_size, num_classes, num_dense_layers = 6, dropout_p=0.0):
        super(MLP, self).__init__()
        self.dense_blocks = []
        self.dense_blocks.append(DenseBlock(input_size, hidden_size*2,dropout_p=0.0))
        self.dense_blocks.append(DenseBlock(hidden_size*2, hidden_size*3,dropout_p=0.0))
        self.dense_blocks.append(DenseBlock(hidden_size*3, hidden_size,dropout_p=0.0))
        for i in range(num_dense_layers-1):
            self.dense_blocks.append(DenseBlock(hidden_size, hidden_size,dropout_p=0.0))
        self.dense_blocks.append(DenseBlock(hidden_size, num_classes, last_layer=True,dropout_p=dropout_p))
        self.dense_blocks = nn.Sequential(*self.dense_blocks)
        # self.test_a = ResidualBlock1D(1,32)
        # self.test_b = ResidualBlock1D(32,128)
        # self.test_c = ResidualBlock1D(128,64)
        # self.test_d = ResidualBlock1D(64,8)
        # self.test_e = nn.Linear(1024,1)

    def forward(self, x):
        x = x.squeeze()
        out = self.dense_blocks(x)

        # out = self.test_a(x)
        # out = self.test_b(out)
        # out = self.test_c(out)
        # out = self.test_d(out)
        # out = out.flatten(start_dim=1)
        # out = self.test_e(out)

        return out


class ResidualAttentionModel_92(nn.Module):
    # for input size 224
    def __init__(self, output_dim_ffe, dropout=False):
        super(ResidualAttentionModel_92, self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias = False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True)
        )
        self.dropout = dropout
        self.mpool1 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.ResidualBlock1 = ResidualBlock(64, 256)
        if self.dropout:
            self.dp_1 = nn.Dropout(0.2)
            self.dp_2 = nn.Dropout(0.2)
            self.dp_3 = nn.Dropout(0.2)
        self.attention_module1 = AttentionModule_stage1(256, 256)
        self.ResidualBlock2 = ResidualBlock(256, 512, 2)
        self.attention_module2 = AttentionModule_stage2(512, 512)
        self.attention_module2_2 = AttentionModule_stage2(512, 512)  # tbq add
        self.ResidualBlock3 = ResidualBlock(512, 1024, 2)
        self.attention_module3 = AttentionModule_stage3(1024, 1024)
        self.attention_module3_2 = AttentionModule_stage3(1024, 1024)  # tbq add
        self.attention_module3_3 = AttentionModule_stage3(1024, 1024)  # tbq add
        self.ResidualBlock4 = ResidualBlock(1024, 2048, 2)
        self.ResidualBlock5 = ResidualBlock(2048, 2048)
        self.ResidualBlock6 = ResidualBlock(2048, 2048)
        self.mpool2 = nn.Sequential(
            nn.BatchNorm2d(2048),
            nn.ReLU(inplace=True),
            nn.AvgPool2d(kernel_size=7, stride=1)
        )
        self.fc_1 = nn.Linear(2048, output_dim_ffe)
        #self.fc_2 = nn.Linear(2048, output_dim_nos)
        #self.softmax = nn.Softmax(dim=1)

        for m in self.children():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_uniform_(m.weight.data)
                nn.init.xavier_uniforml_(m.bias.data)

    def forward(self, x):
        out = self.conv1(x)
        out = self.mpool1(out)
        if self.dropout:
            out = self.dp_1(out)

        out = self.ResidualBlock1(out)
        out = self.attention_module1(out)
        out = self.ResidualBlock2(out)
        out = self.attention_module2(out)
        out = self.attention_module2_2(out)
        if self.dropout:
            out = self.dp_2(out)
        out = self.ResidualBlock3(out)
                # print(out.data)
        out = self.attention_module3(out)
        out = self.attention_module3_2(out)
        out = self.attention_module3_3(out)
        out = self.ResidualBlock4(out)
        out = self.ResidualBlock5(out)
        out = self.ResidualBlock6(out)
        out = self.mpool2(out)
        out = out.view(out.size(0), -1)
        if self.dropout:
            out = self.dp_3(out)

        out_ffe = self.fc_1(out)
        #out_ffe = torch.nn.functional.normalize(out_ffe, 2) # For npid
        #out_nos = self.fc_2(out)
        # Not necessary actually
        #out = self.softmax(out)


        return out_ffe#, out_nos

class ResidualAttentionModel_92_Small(nn.Module):
    # for input size 224
    def __init__(self,output_dim=6, dropout=False):
        super(ResidualAttentionModel_92_Small, self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias = False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True)
        )
        self.dropout = dropout
        self.mpool1 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.ResidualBlock1 = ResidualBlock(64, 256)
        if self.dropout:
            self.dp_1 = nn.Dropout(0.2)
            self.dp_2 = nn.Dropout(0.0)
            self.dp_3 = nn.Dropout(0.0)
        self.attention_module1 = AttentionModule_stage1(256, 256)
        self.ResidualBlock2 = ResidualBlock(256, 512, 2)
        self.attention_module2 = AttentionModule_stage2(512, 512)
        self.attention_module2_2 = AttentionModule_stage2(512, 512)  # tbq add
        self.ResidualBlock3 = ResidualBlock(512, 1024, 3)

        self.mpool2 = nn.Sequential(
            nn.BatchNorm2d(1024),
            nn.ReLU(inplace=True),
            nn.AvgPool2d(kernel_size=8, stride=3)
        )
        self.fc = nn.Linear(1024,output_dim)
        #self.softmax = nn.Softmax(dim=1)

        for m in self.children():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_uniform_(m.weight.data)
                nn.init.xavier_uniforml_(m.bias.data)

    def forward(self, x):
        out = self.conv1(x)
        out = self.mpool1(out)
        if self.dropout:
            out = self.dp_1(out)
        # print(out.data)
        out = self.ResidualBlock1(out)
        out = self.attention_module1(out)
        out = self.ResidualBlock2(out)
        out = self.attention_module2(out)
        out = self.attention_module2_2(out)
        if self.dropout:
            out = self.dp_2(out)
        out = self.ResidualBlock3(out)
                # print(out.data)

        out = self.mpool2(out)
        #print(out.shape)
        out = out.view(out.size(0), -1)
        if self.dropout:
            out = self.dp_3(out)

        out = self.fc(out)
        #out = torch.nn.functional.normalize(out, 2) # For npid
        # Not necessary actually
        #out = self.softmax(out)


        return out

class ResidualAttentionModel_92_Small_1D(nn.Module):
    # for input size 224
    def __init__(self,output_dim=6, dropout=False):
        start_size = 64 # 256
        mid_size = 64 # 512
        end_size = 64 # 1024

        super(ResidualAttentionModel_92_Small_1D, self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv1d(1, 64, kernel_size=2, stride=2, bias = False),
            nn.BatchNorm1d(64),
            nn.ReLU(inplace=True)
        )
        self.dropout = dropout
        self.mpool1 = nn.MaxPool1d(kernel_size=3, stride=2, padding=1)
        self.ResidualBlock1 = ResidualBlock1D(64, start_size)
        if self.dropout:
            self.dp_1 = nn.Dropout(0.4)
            self.dp_2 = nn.Dropout(0.4)
            self.dp_3 = nn.Dropout(0.4)
        #self.attention_module1 = AttentionModule_stage1(256, 256)
        self.ResidualBlock2 = ResidualBlock1D(start_size, mid_size, 2)
        #self.attention_module2 = AttentionModule_stage2(512, 512)
        #self.attention_module2_2 = AttentionModule_stage2(512, 512)  # tbq add

        self.ResidualBlock3 = ResidualBlock1D(mid_size, end_size, 3)

        self.mpool2 = nn.Sequential(
            nn.BatchNorm1d(end_size),
            nn.ReLU(inplace=True),
            nn.AvgPool1d(kernel_size=4, stride=3)
        )
        self.fc = nn.Linear(end_size,output_dim)
        #self.softmax = nn.Softmax(dim=1)

        for m in self.children():
            if isinstance(m, nn.Conv1d):
                nn.init.xavier_uniform_(m.weight.data)
                nn.init.xavier_uniforml_(m.bias.data)

    def forward(self, x):
        out = self.conv1(x)
        out = self.mpool1(out)
        if self.dropout:
            out = self.dp_1(out)
        # print(out.data)
        out = self.ResidualBlock1(out)
        #out = self.attention_module1(out)
        out = self.ResidualBlock2(out)
        #out = self.attention_module2(out)
        #out = self.attention_module2_2(out)
        if self.dropout:
            out = self.dp_2(out)
        out = self.ResidualBlock3(out)
                # print(out.data)

        out = self.mpool2(out)
        #print(out.shape)
        out = out.view(out.size(0), -1)
        if self.dropout:
            out = self.dp_3(out)

        out = self.fc(out)
        # Not necessary actually
        #out = self.softmax(out)


        return out


class ResidualAttentionModel_92_32input(nn.Module):
    # for input size 32
    def __init__(self, output_dim):
        super(ResidualAttentionModel_92_32input, self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 32, kernel_size=5, stride=1, padding=2, bias=False),
            nn.BatchNorm2d(32),
            nn.ReLU(inplace=True)
        )  # 32*32
        self.mpool1 = nn.MaxPool2d(kernel_size=2, stride=2, padding=0)  # 16*16
        self.ResidualBlock1 = ResidualBlock(32, 128)  # 16*16
        self.attention_module1 = AttentionModule_stage1_cifar(128, 128)  # 16*16
        self.ResidualBlock2 = ResidualBlock(128, 256, 2)  # 8*8
        self.attention_module2 = AttentionModule_stage2_cifar(256, 256)  # 8*8
        self.attention_module2_2 = AttentionModule_stage2_cifar(256, 256)  # 8*8 # tbq add
        self.ResidualBlock3 = ResidualBlock(256, 512, 2)  # 4*4
        self.attention_module3 = AttentionModule_stage3_cifar(512, 512)  # 4*4
        self.attention_module3_2 = AttentionModule_stage3_cifar(512, 512)  # 4*4 # tbq add
        self.attention_module3_3 = AttentionModule_stage3_cifar(512, 512)  # 4*4 # tbq add
        self.ResidualBlock4 = ResidualBlock(512, 1024)  # 4*4
        self.ResidualBlock5 = ResidualBlock(1024, 1024)  # 4*4
        self.ResidualBlock6 = ResidualBlock(1024, 1024)  # 4*4
        self.mpool2 = nn.Sequential(
            nn.BatchNorm2d(1024),
            nn.ReLU(inplace=True),
            nn.AvgPool2d(kernel_size=4, stride=1)
        )
        self.fc = nn.Linear(1024,output_dim)

    def forward(self, x):
        out = self.conv1(x)
        out = self.mpool1(out)
        # print(out.data)
        out = self.ResidualBlock1(out)
        out = self.attention_module1(out)
        out = self.ResidualBlock2(out)
        out = self.attention_module2(out)
        out = self.attention_module2_2(out)
        out = self.ResidualBlock3(out)
        # print(out.data)
        out = self.attention_module3(out)
        out = self.attention_module3_2(out)
        out = self.attention_module3_3(out)
        out = self.ResidualBlock4(out)
        out = self.ResidualBlock5(out)
        out = self.ResidualBlock6(out)
        out = self.mpool2(out)
        out = out.view(out.size(0), -1)
        out = self.fc(out)

        return out


class ResidualAttentionModel_92_32input_update(nn.Module):
    # for input size 32
    def __init__(self, output_dim):
        super(ResidualAttentionModel_92_32input_update, self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 32, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(32),
            nn.ReLU(inplace=True)
        )  # 32*32
        # self.mpool1 = nn.MaxPool2d(kernel_size=2, stride=2, padding=0)  # 16*16
        self.ResidualBlock1 = ResidualBlock(32, 128)  # 32*32
        self.attention_module1 = AttentionModule_stage1_cifar(128, 128, size1=(32, 32), size2=(16, 16))  # 32*32
        self.ResidualBlock2 = ResidualBlock(128, 256, 2)  # 16*16
        self.attention_module2 = AttentionModule_stage2_cifar(256, 256, size=(16, 16))  # 16*16
        self.attention_module2_2 = AttentionModule_stage2_cifar(256, 256, size=(16, 16))  # 16*16 # tbq add
        self.ResidualBlock3 = ResidualBlock(256, 512, 2)  # 4*4
        self.attention_module3 = AttentionModule_stage3_cifar(512, 512)  # 8*8
        self.attention_module3_2 = AttentionModule_stage3_cifar(512, 512)  # 8*8 # tbq add
        self.attention_module3_3 = AttentionModule_stage3_cifar(512, 512)  # 8*8 # tbq add
        self.ResidualBlock4 = ResidualBlock(512, 1024)  # 8*8
        self.ResidualBlock5 = ResidualBlock(1024, 1024)  # 8*8
        self.ResidualBlock6 = ResidualBlock(1024, 1024)  # 8*8
        self.mpool2 = nn.Sequential(
            nn.BatchNorm2d(1024),
            nn.ReLU(inplace=True),
            nn.AvgPool2d(kernel_size=8)
        )
        self.fc = nn.Linear(1024, output_dim)

    def forward(self, x):
        out = self.conv1(x)
        # out = self.mpool1(out)
        # print(out.data)
        out = self.ResidualBlock1(out)
        out = self.attention_module1(out)
        out = self.ResidualBlock2(out)
        out = self.attention_module2(out)
        out = self.attention_module2_2(out)
        out = self.ResidualBlock3(out)
        # print(out.data)
        out = self.attention_module3(out)
        out = self.attention_module3_2(out)
        out = self.attention_module3_3(out)
        out = self.ResidualBlock4(out)
        out = self.ResidualBlock5(out)
        out = self.ResidualBlock6(out)
        out = self.mpool2(out)
        out = out.view(out.size(0), -1)
        out = self.fc(out)

        return out


'''ResNet in PyTorch.

For Pre-activation ResNet, see 'preact_resnet.py'.

Reference:
[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
    Deep Residual Learning for Image Recognition. arXiv:1512.03385
'''
import torch
import torch.nn as nn
import torch.nn.functional as F
from lib.normalize import Normalize

from torch.autograd import Variable


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion*planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion*planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, low_dim=128):
        super(ResNet, self).__init__()
        self.in_planes = 64

        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2)
        self.linear = nn.Linear(512*block.expansion, low_dim)
        self.l2norm = Normalize(2)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = F.avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        out = self.l2norm(out)
        return out


def ResNet18(low_dim=128):
    return ResNet(BasicBlock, [2,2,2,2], low_dim)

def ResNet34(low_dim=128):
    return ResNet(BasicBlock, [3,4,6,3], low_dim)

def ResNet50(low_dim=128):
    return ResNet(Bottleneck, [3,4,6,3], low_dim)

def ResNet101(low_dim=128):
    return ResNet(Bottleneck, [3,4,23,3], low_dim)

def ResNet152(low_dim=128):
    return ResNet(Bottleneck, [3,8,36,3], low_dim)


def test():
    net = ResNet18()
    y = net(Variable(torch.randn(1,3,32,32)))
    print(y.size())

# test()



import sys
import os
import logging
import re
import functools
import fnmatch
import numpy as np


def setup_logger(distributed_rank=0, filename="log.txt"):
    logger = logging.getLogger("Logger")
    logger.setLevel(logging.DEBUG)
    # don't log results for the non-master process
    if distributed_rank > 0:
        return logger
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.DEBUG)
    fmt = "[%(asctime)s %(levelname)s %(filename)s line %(lineno)d %(process)d] %(message)s"
    ch.setFormatter(logging.Formatter(fmt))
    logger.addHandler(ch)

    return logger


def find_recursive(root_dir, ext='.jpg'):
    files = []
    for root, dirnames, filenames in os.walk(root_dir):
        for filename in fnmatch.filter(filenames, '*' + ext):
            files.append(os.path.join(root, filename))
    return files


class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self):
        self.initialized = False
        self.val = None
        self.avg = None
        self.sum = None
        self.count = None

    def initialize(self, val, weight):
        self.val = val
        self.avg = val
        self.sum = val * weight
        self.count = weight
        self.initialized = True

    def update(self, val, weight=1):
        if not self.initialized:
            self.initialize(val, weight)
        else:
            self.add(val, weight)

    def add(self, val, weight):
        self.val = val
        self.sum += val * weight
        self.count += weight
        self.avg = self.sum / self.count

    def value(self):
        return self.val

    def average(self):
        return self.avg


def unique(ar, return_index=False, return_inverse=False, return_counts=False):
    ar = np.asanyarray(ar).flatten()

    optional_indices = return_index or return_inverse
    optional_returns = optional_indices or return_counts

    if ar.size == 0:
        if not optional_returns:
            ret = ar
        else:
            ret = (ar,)
            if return_index:
                ret += (np.empty(0, np.bool),)
            if return_inverse:
                ret += (np.empty(0, np.bool),)
            if return_counts:
                ret += (np.empty(0, np.intp),)
        return ret
    if optional_indices:
        perm = ar.argsort(kind='mergesort' if return_index else 'quicksort')
        aux = ar[perm]
    else:
        ar.sort()
        aux = ar
    flag = np.concatenate(([True], aux[1:] != aux[:-1]))

    if not optional_returns:
        ret = aux[flag]
    else:
        ret = (aux[flag],)
        if return_index:
            ret += (perm[flag],)
        if return_inverse:
            iflag = np.cumsum(flag) - 1
            inv_idx = np.empty(ar.shape, dtype=np.intp)
            inv_idx[perm] = iflag
            ret += (inv_idx,)
        if return_counts:
            idx = np.concatenate(np.nonzero(flag) + ([ar.size],))
            ret += (np.diff(idx),)
    return ret


def colorEncode(labelmap, colors, mode='RGB'):
    labelmap = labelmap.astype('int')
    labelmap_rgb = np.zeros((labelmap.shape[0], labelmap.shape[1], 3),
                            dtype=np.uint8)
    for label in unique(labelmap):
        if label < 0:
            continue
        labelmap_rgb += (labelmap == label)[:, :, np.newaxis] * \
            np.tile(colors[label],
                    (labelmap.shape[0], labelmap.shape[1], 1))

    if mode == 'BGR':
        return labelmap_rgb[:, :, ::-1]
    else:
        return labelmap_rgb


def accuracy(preds, label):
    valid = (label >= 0)
    acc_sum = (valid * (preds == label)).sum()
    valid_sum = valid.sum()
    acc = float(acc_sum) / (valid_sum + 1e-10)
    return acc, valid_sum


def intersectionAndUnion(imPred, imLab, numClass):
    imPred = np.asarray(imPred).copy()
    imLab = np.asarray(imLab).copy()

    imPred += 1
    imLab += 1
    # Remove classes from unlabeled pixels in gt image.
    # We should not penalize detections in unlabeled portions of the image.
    imPred = imPred * (imLab > 0)

    # Compute area intersection:
    intersection = imPred * (imPred == imLab)
    (area_intersection, _) = np.histogram(
        intersection, bins=numClass, range=(1, numClass))

    # Compute area union:
    (area_pred, _) = np.histogram(imPred, bins=numClass, range=(1, numClass))
    (area_lab, _) = np.histogram(imLab, bins=numClass, range=(1, numClass))
    area_union = area_pred + area_lab - area_intersection

    return (area_intersection, area_union)


class NotSupportedCliException(Exception):
    pass


def process_range(xpu, inp):
    start, end = map(int, inp)
    if start > end:
        end, start = start, end
    return map(lambda x: '{}{}'.format(xpu, x), range(start, end+1))


REGEX = [
    (re.compile(r'^gpu(\d+)$'), lambda x: ['gpu%s' % x[0]]),
    (re.compile(r'^(\d+)$'), lambda x: ['gpu%s' % x[0]]),
    (re.compile(r'^gpu(\d+)-(?:gpu)?(\d+)$'),
     functools.partial(process_range, 'gpu')),
    (re.compile(r'^(\d+)-(\d+)$'),
     functools.partial(process_range, 'gpu')),
]


def parse_devices(input_devices):

    """Parse user's devices input str to standard format.
    e.g. [gpu0, gpu1, ...]

    """
    ret = []
    for d in input_devices.split(','):
        for regex, func in REGEX:
            m = regex.match(d.lower().strip())
            if m:
                tmp = func(m.groups())
                # prevent duplicate
                for x in tmp:
                    if x not in ret:
                        ret.append(x)
                break
        else:
            raise NotSupportedCliException(
                'Can not recognize device: "{}"'.format(d))
    return ret



from setuptools import setup,find_packages
from os import path as os_path

import brails

this_directory = os_path.abspath(os_path.dirname(__file__))


def read_file(filename):
    with open(os_path.join(this_directory, filename), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


def read_requirements(filename):
    return [line.strip() for line in read_file(filename).splitlines()
            if not line.startswith('#')]


setup(
    name='BRAILS',
    python_requires='>=3.6',
    version=brails.__version__,
    description="Building Recognition using AI at Large-Scale",
    long_description=read_file('README.md'),
    long_description_content_type="text/markdown",
    author="Charles Wang",
    author_email='c_w@berkeley.edu',
    url='https://github.com/NHERI-SimCenter/BRAILS',
    #packages=['brails'],
    packages=find_packages(),
    zip_safe=False,
    install_requires=read_requirements('requirements.txt'),
    include_package_data=True,
    package_data={'': ['brails/modules/Foundation_Classification/csail_segmentation_tool/csail_seg/data/color150.mat']},
    license="BSD 3-Clause",
    keywords=['brails', 'bim', 'brails framework'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



import torch
import numpy as np

class AliasMethod(object):
    '''
        From: https://hips.seas.harvard.edu/blog/2013/03/03/the-alias-method-efficient-sampling-with-many-discrete-outcomes/
    '''
    def __init__(self, probs):

        if probs.sum() > 1:
            probs.div_(probs.sum())
        K = len(probs)
        self.prob = torch.zeros(K)
        self.alias = torch.LongTensor([0]*K)

        # Sort the data into the outcomes with probabilities
        # that are larger and smaller than 1/K.
        smaller = []
        larger = []
        for kk, prob in enumerate(probs):
            self.prob[kk] = K*prob
            if self.prob[kk] < 1.0:
                smaller.append(kk)
            else:
                larger.append(kk)

        # Loop though and create little binary mixtures that
        # appropriately allocate the larger outcomes over the
        # overall uniform mixture.
        while len(smaller) > 0 and len(larger) > 0:
            small = smaller.pop()
            large = larger.pop()

            self.alias[small] = large
            self.prob[large] = (self.prob[large] - 1.0) + self.prob[small]

            if self.prob[large] < 1.0:
                smaller.append(large)
            else:
                larger.append(large)

        for last_one in smaller+larger:
            self.prob[last_one] = 1

    def cuda(self): 
        self.prob = self.prob.cuda()
        self.alias = self.alias.cuda()

    def draw(self, N):
        '''
            Draw N samples from multinomial
        '''
        K = self.alias.size(0)

        kk = torch.zeros(N, dtype=torch.long, device=self.prob.device).random_(0, K)
        prob = self.prob.index_select(0, kk)
        alias = self.alias.index_select(0, kk)
        # b is whether a random number is greater than q
        b = torch.bernoulli(prob)
        oq = kk.mul(b.long())
        oj = alias.mul((1-b).long())

        return oq + oj




# -*- coding: utf-8 -*-
"""
/*------------------------------------------------------*
|                         BRAILS                        |
|                                                       |
| Author: Charles Wang,  UC Berkeley, c_w@berkeley.edu  |
|                                                       |
| Date:    10/15/2020                                   |
*------------------------------------------------------*/
"""


zoo = {
    'roofType':{
        'fileURL' : 'https://zenodo.org/record/4059084/files/roof_classifier_v0.1.h5',
        'classNames' : ['flat', 'gabled', 'hipped']
    } ,
    'residentialOccupancyClass':{
        'fileURL' : 'https://zenodo.org/record/4091547/files/occupancy-78-78-79.h5',
        'classNames' : ['RES1', 'RES3']
    } ,
    'occupancyClass':{
        'fileURL' : 'https://zenodo.org/record/4553798/files/occupancy_InceptionV3_V0.2.h5',
        'classNames' : ['RES3', 'COM' ,'RES1']
    } ,
    'softstory':{
        'fileURL' : 'https://zenodo.org/record/4094334/files/softstory-80-81-87.5-v0.1.h5',
        'classNames' : ['others', 'softstory']
    } 
}


import argparse

import torch
import torchvision.transforms as transforms
from torch.utils.data import DataLoader
import os
import csv

from PIL import Image
from torch.utils.data import Dataset
from torch.autograd.variable import Variable
import sklearn
from sklearn.metrics.classification import precision_recall_fscore_support

import numpy as np
import pandas as pd

import matplotlib
matplotlib.use('agg')

import matplotlib.pyplot as plt
import seaborn as sns
from lib.datasets import YearBuiltFolder

parser = argparse.ArgumentParser(description='Classify Year Built')

parser.add_argument('--image-path',help='Path to one image or a folder containing images.',required=True)
parser.add_argument('--checkpoint', default=None,type=str,
                    help='Path to checkpoint. Defaults to best pretrained version.')
parser.add_argument('--only-cpu', action='store_true', help='Use CPU only, disregard GPU.')

parser.add_argument('--model',help='Pretrained model, options ["foundation_v0.1"]', type=str)
parser.add_argument('--calc-perf', help='Calculate the performance and save a confusion matrix.',action='store_true')


args = parser.parse_args()

if args.only_cpu:
    device='cpu'
else:
    device = 'cuda' if torch.cuda.is_available() else 'cpu'

def main():
    
    test_transforms = transforms.Compose([
        transforms.Scale((550, 550)),
        transforms.CenterCrop(448),
        transforms.ToTensor(),
        transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
    ])    
      
    if args.checkpoint is not None:
        model = torch.load(args.checkpoint)
    else:
        weight_file_path = './weights/model_weights.pth'
        if not os.path.isfile(weight_file_path):
            print('Loading remote model file to the weights folder..')
            torch.hub.download_url_to_file('https://zenodo.org/record/4310463/files/model_best.pth', weight_file_path)
            
        model = torch.load(weight_file_path)

    try:
        num_classes = model.classifier1[4].out_features # Take amount of classes from model file
    except AttributeError:
        raise RuntimeError('The model, either provided or downloaded, does not fit the format in this implementation. Please check if the checkpoint is correct and fits to the model.')
    
    dataset = YearBuiltFolder(args.image_path,  transforms=test_transforms, classes=range(num_classes),calc_perf = args.calc_perf)
    # Do not change the batchsize.
    test_loader = DataLoader(dataset, batch_size=1, shuffle=False, num_workers=0) 

    model.eval()
    model = model.to(device)
    
    print('Checking {} images'.format(len(dataset)))

    predictions = evaluate_to_stats(model, test_loader)

    with open('{}_prediction_results.csv'.format(os.path.basename(os.path.normpath(args.image_path))), 'w', newline='') as myfile:
        wr = csv.writer(myfile, quoting=csv.QUOTE_ALL)
        if args.calc_perf:
            wr.writerow(['filename','prediction','ground truth'])
        else:
            wr.writerow(['filename','prediction'])
            
        for prediction in predictions:
            wr = csv.writer(myfile, quoting=csv.QUOTE_ALL)
            if args.calc_perf:
                wr.writerow([str(prediction['filename']),str(prediction['prediction'][0]),str(prediction['ground truth'][0])])
            else:
                wr.writerow([str(prediction['filename']),str(prediction['prediction'][0])])

    print ('Classification finished. Results written to {}'.format('{}_prediction_results.csv'.format(os.path.basename(os.path.normpath(args.image_path)))))
    
def evaluate_to_stats(net, testloader):
    net.eval()
    
    predictions = []
    
    with torch.no_grad():
        for batch_idx, (inputs, label, filename) in enumerate(testloader):
    
            if inputs.shape[0] > 1:
                raise NotImplementedError('Please choose a batch size of 1. Saving the results is not compatible with larger batch sizes in this version.')        
            inputs = inputs.to(device)
            inputs = Variable(inputs)
            output_1, output_2, output_3, output_concat= net(inputs)
            outputs_com = output_1 + output_2 + output_3 + output_concat
    
            _, predicted_com = torch.max(outputs_com.data, 1)
            
            y_pred = predicted_com[0].flatten().cpu().numpy()
            
            y_fn = filename[0]
            
            if args.calc_perf:
                predictions.append({'filename':y_fn,'prediction':y_pred, 'ground truth':label.cpu().numpy()})
            else:
                predictions.append({'filename':y_fn,'prediction':y_pred})
                
               
            if batch_idx % 50 == 0:
                print('Testing image {} from {}'.format(batch_idx,len(testloader)))
                
        if args.calc_perf:
            y_gt = []
            y_pred = []
            for prediction in predictions:
                y_gt.append(prediction['ground truth'])
                y_pred.append(prediction['prediction'])
            
            y_gt = np.array(y_gt,dtype=np.uint8)
            y_pred = np.array(y_pred,dtype=np.uint8)
            
            precision, recall, f1, support = precision_recall_fscore_support(y_gt,
                                                                         y_pred, average='macro')
                    
            confusion_matrix = sklearn.metrics.confusion_matrix(y_gt, y_pred, labels=range(len(testloader.dataset.classes)))
            
            cm_fig = construct_confusion_matrix_image(testloader.dataset.classes, confusion_matrix)
        
            cm_fig.savefig('result_confusion_matrix.png',dpi=300)
            
            print('F1 {}, precision, {}, recall, {}'.format(f1,precision,recall))   

    return predictions

def construct_confusion_matrix_image(classes, con_mat):
    con_mat_norm = np.around(con_mat.astype('float') / con_mat.sum(axis=1)[:, np.newaxis], decimals=2)

    con_mat_df = pd.DataFrame(con_mat_norm,
                              index=classes,
                              columns=classes)

    figure = plt.figure(figsize=(8, 8))
    sns.heatmap(con_mat_df, annot=True, cmap=plt.cm.Blues)
    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')
    figure.canvas.draw()

    return figure


if __name__ == '__main__':
    main()


# Author: Zylo117

import math

from torch import nn
import torch.nn.functional as F


class Conv2dStaticSamePadding(nn.Module):
    """
    created by Zylo117
    The real keras/tensorflow conv2d with same padding
    """

    def __init__(self, in_channels, out_channels, kernel_size, stride=1, bias=True, groups=1, dilation=1, **kwargs):
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride=stride,
                              bias=bias, groups=groups)
        self.stride = self.conv.stride
        self.kernel_size = self.conv.kernel_size
        self.dilation = self.conv.dilation

        if isinstance(self.stride, int):
            self.stride = [self.stride] * 2
        elif len(self.stride) == 1:
            self.stride = [self.stride[0]] * 2

        if isinstance(self.kernel_size, int):
            self.kernel_size = [self.kernel_size] * 2
        elif len(self.kernel_size) == 1:
            self.kernel_size = [self.kernel_size[0]] * 2

    def forward(self, x):
        h, w = x.shape[-2:]
        
        extra_h = (math.ceil(w / self.stride[1]) - 1) * self.stride[1] - w + self.kernel_size[1]
        extra_v = (math.ceil(h / self.stride[0]) - 1) * self.stride[0] - h + self.kernel_size[0]
        
        left = extra_h // 2
        right = extra_h - left
        top = extra_v // 2
        bottom = extra_v - top

        x = F.pad(x, [left, right, top, bottom])

        x = self.conv(x)
        return x


class MaxPool2dStaticSamePadding(nn.Module):
    """
    created by Zylo117
    The real keras/tensorflow MaxPool2d with same padding
    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.pool = nn.MaxPool2d(*args, **kwargs)
        self.stride = self.pool.stride
        self.kernel_size = self.pool.kernel_size

        if isinstance(self.stride, int):
            self.stride = [self.stride] * 2
        elif len(self.stride) == 1:
            self.stride = [self.stride[0]] * 2

        if isinstance(self.kernel_size, int):
            self.kernel_size = [self.kernel_size] * 2
        elif len(self.kernel_size) == 1:
            self.kernel_size = [self.kernel_size[0]] * 2

    def forward(self, x):
        h, w = x.shape[-2:]
        
        extra_h = (math.ceil(w / self.stride[1]) - 1) * self.stride[1] - w + self.kernel_size[1]
        extra_v = (math.ceil(h / self.stride[0]) - 1) * self.stride[0] - h + self.kernel_size[0]

        left = extra_h // 2
        right = extra_h - left
        top = extra_v // 2
        bottom = extra_v - top

        x = F.pad(x, [left, right, top, bottom])

        x = self.pool(x)
        return x



import json 
import xmltodict 
import glob
import os
import subprocess
import numpy as np
import xml.etree.ElementTree as ET
import pandas as pd
from os import path
import json
import random
from detectron2.structures import BoxMode
from detectron2.data.datasets import register_coco_instances
import pandas as pd

# import some common libraries
import numpy as np
import cv2
import random
from cv2 import imshow

# import some common detectron2 utilities
from detectron2 import model_zoo
from detectron2.engine import DefaultPredictor
from detectron2.config import get_cfg
from detectron2.utils.visualizer import Visualizer
from detectron2.data import MetadataCatalog
import matplotlib.pyplot as plt



from detectron2.data.datasets import register_coco_instances
from detectron2.engine import DefaultTrainer

import glob
import random
from detectron2.structures import BoxMode
from detectron2.data.catalog import DatasetCatalog

def encodeClass(df,className,fileName,fileNameReverse):
    classEncodes={}
    classEncodes_reverse={}
    for i,c in enumerate(list(df[className].unique())):
        classEncodes[c] = i+1
        classEncodes_reverse[i+1]=c
    print(classEncodes)
    with open(fileName, 'w') as fp:
        json.dump(classEncodes, fp)
    with open(fileNameReverse, 'w') as fp:
        json.dump(classEncodes_reverse, fp)
        
def writeanno(df, filename,tagName,classEncodeFile,imgDir='../'):
    images = []
    annotations = []
    
    
    imgID = 0
    imgList = []
    
    with open(classEncodeFile) as json_file:
        classEncode = json.load(json_file)
    
    for index, row in df.iterrows():

        category = row[tagName]
        #if category == "unknown":
        #    continue

        #categoryID = 1 if category=="reinforced" else 2
        categoryID = classEncode[category]

        w = int(row['width'])
        h = int(row['height'])

        #img_name = row['imgPath']
        #FolderID = row['FolderID']
        #imgPath = "/Users/simcenter/Files/Work/WorldBank/DRONE/SALINACRUZ/{}/{}".format(FolderID, img_name)
        imgPath = row['imgPath']
        imgName_new = imgPath
        #imgName_new = category + "/" + str(index+1)+".jpg"

        print(imgDir+imgPath)
        
        if not path.exists(imgDir+imgPath):
            continue     
            

        x1 = int(round(float(row['xtl'])))
        y1 = int(round(float(row['ytl'])))
        x2 = int(round(float(row['xbr'])))
        y2 = int(round(float(row['ybr'])))
        
  

        if imgPath not in imgList:
            imgList.append(imgPath)
            imgID += 1
            image = {"id" : imgID,
                    "file_name" : imgName_new,
                    "width": w,
                    "height": h,
                    "date_captured": "2020-10-25 01:26:40.419787",
                    "license": 1,
                    "coco_url": "",
                    "flickr_url": ""}
            images.append(image)
            currentImgID = imgID
        else: currentImgID = imgList.index(imgPath)+1





        try:

            segmentation =  [[x1,y1, x2,y1, x2,y2, x1,y2]]

            item = {
                "area" : w*h,
                "segmentation" : segmentation,
                "iscrowd" : 0,
                "image_id" : currentImgID,
                "bbox" : [x1,y1,x2-x1,y2-y1],
                'bbox_mode' : 4, # BoxMode.XYWHA_ABS
                
                "category_id" : categoryID,
                "id" : index+1
                #"size" : imgSize,
                #"name" : imgName_new,
                #"filename" : imgName_new,
                #"regions" : regions
            }
            annotations.append(item)
            
            #shutil.copyfile(imgPath, newDir+imgName_new) 
            
        except:
            #print(imgPath+' doesn\'t exit')
            #cmd = 'aws s3 cp s3://wbg-geography01/GEP/DRONE/SALINACRUZ/{}/{} {}\n'.format(FolderID,img_name, imgPath)
            #print(cmd)
            pass #print(imgPath)

    labels = {}
    labels["annotations"] = annotations
    labels["images"] = images
    labels["info"] = {
    "description": "Building Dataset",
    "url": ".",
    "version": "0.1.0",
    "year": 2020,
    "contributor": "charles",
    "date_created": "2020-04-25 01:29:34.597570"
    }
    
    labels["licenses"] = [
    {
    "id": 1,
    "name": ".",
    "url": "."
    }
    ]
    '''
    labels["categories"] = [
    {
    "id": 1,
    "name": "reinforced",
    "supercategory": "building"
    },
    {
    "id": 2,
    "name": "unreinforced",
    "supercategory": "building"
    }
    ]
    '''
    
    labels_categories = []
    for classN in classEncode.keys():
        c = {
            'id': classEncode[classN],
            'name': classN,
            'supercategory': 'building'
        }
        labels_categories.append(c)
    labels["categories"] = labels_categories

    with open(filename, 'w') as outfile:
        json.dump(labels, outfile)
    
    print(filename)
 


    
def downloadimages(df):

    for index, row in df.iterrows():

        imgPath = row['imgPath']

        # download from s3
        if not path.exists('../'+imgPath):
            cmd = 'aws s3 cp s3://wbg-geography01/GEP/'+imgPath+' ../'+imgPath
            os.system(cmd)
            print(' ../'+imgPath)


        

def getImgbyID(id,j):
  for img in j['images']:
    if img['id'] == id:
      return img['file_name'], img['height'],img['width']
  assert "didn't find file..."
    
def get_dataset_dicts(annofilename,cat='',imgDir='../'):

    #if name=='all': annofilename='labels_json/{}_test.json'.format(cat)
    #else: annofilename='labels_json/{}_test_{}.json'.format(cat,name)
    with open(annofilename) as f:
      j = json.load(f)
    
    #

    dataset_dicts = []
    for anno in j['annotations']:
      #print(anno)
      objs = []
      obj = {
                    "bbox": anno['bbox'],
                    "bbox_mode": BoxMode.XYWH_ABS,
                    "segmentation": anno['segmentation'],
                    "category_id": j['categories'][anno["category_id"]-1]['name'], #anno["category_id"],
                    "iscrowd": anno['iscrowd']
                }
      objs.append(obj)
      record = {}
      imgID = anno['image_id']
      imgName,h,w = getImgbyID(imgID,j)
      record["file_name"] = imgDir+imgName
      record["image_id"] = imgID
      record["height"] = h
      record["width"] = w
      record["annotations"] = objs
      dataset_dicts.append(record)
    return dataset_dicts


def plotfigs(dataset_dicts,name):
    plt.figure(figsize=(20, 20))
    for i,d in enumerate(random.sample(dataset_dicts, 9)):
        img = cv2.imread(d["file_name"])
        visualizer = Visualizer(img[:, :, ::-1], metadata=MetadataCatalog.get("checkLabels{}".format(name)), scale=1.0)
        visualizer._default_font_size = max(np.sqrt(visualizer.output.height * visualizer.output.width) // 90 *8, 10 // 1.0 * 8)
        vis = visualizer.draw_dataset_dict(d)
        
        #cv2_imshow(vis.get_image()[:, :, ::-1])
        ax = plt.subplot(3, 3, i + 1)
        plt.imshow(vis.get_image()[:, :, ::-1])
        plt.title(d["file_name"])
        plt.axis("off")
        
def read_anno(annofilename,cat):
    
    #maxNumPerCat = 100
    
    imgPath=[]

    width=[]
    height=[]
    xtl=[]
    ytl=[]
    xbr=[]
    ybr=[]

    category=[]
    
    #if name=='all': annofilename='labels_json/{}_test.json'.format(cat)
    #else: annofilename='labels_json/{}_test_{}.json'.format(cat,name)
    with open(annofilename) as f:
      j = json.load(f)
    
    numCat = len(j['categories'])
    catNumList = []
    [catNumList.append(0) for i in range(numCat)]
    

    dataset_dicts = []
    ll=j['annotations']
    #random.shuffle(ll) 
    for anno in ll:
      #print(anno)
      thisCatID = anno["category_id"]-1

      #if catNumList[thisCatID]>= maxNumPerCat: continue
      catNumList[thisCatID]=catNumList[thisCatID]+1
    
    
      objs = []
      obj = {
                    "bbox": anno['bbox'],
                    "bbox_mode": BoxMode.XYWH_ABS,
                    "segmentation": anno['segmentation'],
                    "category_id": j['categories'][anno["category_id"]-1]['name'], #anno["category_id"],
                    "iscrowd": anno['iscrowd']
                }
      objs.append(obj)
      record = {}
      imgID = anno['image_id']
      imgName,h,w = getImgbyID(imgID,j)
      
      imgPath.append(imgName)
      a,b,c,d=anno['bbox']

      width.append(w)
      height.append(h)
      xtl.append(a)
      ytl.append(b)
      xbr.append(c+a)
      ybr.append(d+b)
      #material
      category.append(j['categories'][anno["category_id"]-1]['name'])
      #condition
      #completeness
    imgPath
    df_tmpx = pd.DataFrame(list(zip(imgPath, width,height,xtl,ytl,xbr,ybr,category)),columns =['imgPath', 'width','height','xtl','ytl','xbr','ybr',cat]) 
    
    return df_tmpx

def read_anno_backup(name,cat):
    imgPath=[]

    width=[]
    height=[]
    xtl=[]
    ytl=[]
    xbr=[]
    ybr=[]

    category=[]
    

    with open('labels_json/{}_test_{}.json'.format(cat,name)) as f:
      j = json.load(f)

    dataset_dicts = []
    for anno in j['annotations']:
      #print(anno)
      objs = []
      obj = {
                    "bbox": anno['bbox'],
                    "bbox_mode": BoxMode.XYWH_ABS,
                    "segmentation": anno['segmentation'],
                    "category_id": j['categories'][anno["category_id"]-1]['name'], #anno["category_id"],
                    "iscrowd": anno['iscrowd']
                }
      objs.append(obj)
      record = {}
      imgID = anno['image_id']
      imgName,h,w = getImgbyID(imgID,j)
      
      imgPath.append(imgName)
      a,b,c,d=anno['bbox']

      width.append(w)
      height.append(h)
      xtl.append(a)
      ytl.append(b)
      xbr.append(c+a)
      ybr.append(d+b)
      #material
      category.append(j['categories'][anno["category_id"]-1]['name'])
      #condition
      #completeness
    imgPath
    df_tmpx = pd.DataFrame(list(zip(imgPath, width,height,xtl,ytl,xbr,ybr,category)),columns =['imgPath', 'width','height','xtl','ytl','xbr','ybr',cat]) 
    
    return df_tmpx


def plotfigs2(dataset_dicts,name,predictor,cfg,fontSize=5,alpha=0.1,threshold=0.7):
    
    
    plt.figure(figsize=(20, 40))
    for i,d in enumerate(random.sample(dataset_dicts, 8)):
        img = cv2.imread(d["file_name"])
        print(d["file_name"])
        #visualizer = Visualizer(img[:, :, ::-1], metadata=MetadataCatalog.get("checkLabels{}".format(name)), scale=1.0)
        visualizer = Visualizer(img[:, :, ::-1], metadata=MetadataCatalog.get('whatever'), scale=1.0)
        visualizer._default_font_size = max(np.sqrt(visualizer.output.height * visualizer.output.width) // 90 *fontSize, 10 // 1.0 * fontSize)
        #print(d)
        #annos = d.get("annotations", None)
        #print(annos)
        #labels = [x["category_id"] for x in annos]

        vis = visualizer.draw_dataset_dict(d,alpha=0.0)
        #cv2_imshow(vis.get_image()[:, :, ::-1])
        ax = plt.subplot(4, 4, 2*i + 1)
        plt.imshow(vis.get_image()[:, :, ::-1])
        #plt.title('Label')#d['annotations'][0]['category_id']
        plt.axis("off")
        
        outputs = predictor(img)
        v = Visualizer(img[:, :, ::-1], MetadataCatalog.get(cfg.DATASETS.TEST[0]), scale=1)
        v._default_font_size = max(np.sqrt(v.output.height * v.output.width) // 90 *fontSize, 10 // 1.2 * fontSize)
        out = v.draw_instance_predictions(outputs["instances"].to("cpu"), alpha=alpha)
        
        ax = plt.subplot(4, 4, 2*i + 2)
        plt.imshow(out.get_image()[:, :, ::-1])
        #plt.title('Prediction')
        plt.axis("off")
        plt.rcParams['axes.xmargin'] = 0

def plotfigs3(dataset_dicts,name,predictor,cfg,cat,fontSize=5,alpha=0.1,threshold=0.7):
    
    if not os.path.exists('images/{}'.format(cat)):
        os.makedirs('images/{}'.format(cat))
    imgDir='images/{}/{}'

    
    for i,d in enumerate(random.sample(dataset_dicts, len(dataset_dicts))):

        print(d["file_name"])

        img = cv2.imread(d["file_name"])

        s = '-'.join(d["file_name"].split('/')[1:])
        newimg=imgDir.format(cat,s)

        if os.path.exists(newimg): continue

        plt.figure(figsize=(20, 10))

        #visualizer = Visualizer(img[:, :, ::-1], metadata=MetadataCatalog.get("checkLabels{}".format(name)), scale=1.0)
        visualizer = Visualizer(img[:, :, ::-1], metadata=MetadataCatalog.get('whatever'), scale=1.0)
        visualizer._default_font_size = max(np.sqrt(visualizer.output.height * visualizer.output.width) // 90 * fontSize, 10 // 1.0 * fontSize)
        #print(d)
        #annos = d.get("annotations", None)
        #print(annos)
        #labels = [x["category_id"] for x in annos]

        vis = visualizer.draw_dataset_dict(d, alpha=0.0)
        #cv2_imshow(vis.get_image()[:, :, ::-1])
        ax = plt.subplot(1,2, 1)
        plt.imshow(vis.get_image()[:, :, ::-1])
        #plt.title('Label')#d['annotations'][0]['category_id']
        plt.axis("off")
        
        outputs = predictor(img)
        v = Visualizer(img[:, :, ::-1], MetadataCatalog.get(cfg.DATASETS.TEST[0]), scale=1)
        v._default_font_size = max(np.sqrt(v.output.height * v.output.width) // 90 *fontSize, 10 // 1.2 * fontSize)
        out = v.draw_instance_predictions(outputs["instances"].to("cpu"), alpha=alpha)
        
        ax = plt.subplot(1,2, 2)
        plt.imshow(out.get_image()[:, :, ::-1])
        #plt.title('Prediction')
        plt.axis("off")
        plt.rcParams['axes.xmargin'] = 0

        plt.savefig(newimg,bbox_inches='tight')



def registerdata(name,testAnno,imgDataDir):
    try:
        register_coco_instances(name, {}, testAnno, imgDataDir)
    except:
        name=name+str(random.randint(1,1e9))
        name=registerdata(name,testAnno,imgDataDir)

    return name

   
def predict(cat,trainjson,testjson,name,testdict,saveimages=False,lr=0.00025,imgDataDir = '../',fontSize=3,alpha=0.1,threshold=0.7,weights=''):
    #trainAnno='labels_json/use_train.json'
    #testAnno='labels_json/use_test_SalinaCruz_test.json'
    #testAnno='labels_json/use_test_Juchitan_test.json'
    trainAnno=trainjson#'labels_json/{}'.format(trainjson)
    testAnno=testjson#'labels_json/{}'.format(testjson)
    
    
    
    try: register_coco_instances("trainset", {}, trainAnno, imgDataDir)
    except: pass
    name = registerdata(name, testAnno, imgDataDir)

    NUM_WORKERS = 2
    IMS_PER_BATCH = 2
    BATCH_SIZE_PER_IMAGE = 128
    rootmodel = 'COCO-InstanceSegmentation'
    modelName='mask_rcnn_R_101_DC5_3x'
    modelPrefix=cat
    
    
    #lr=0.00025
    numiter=120000
    
    if weights=='':
        weights = "model/{}/{}_{}_{}_{}_{}_{}_{}_{}.pth".format(rootmodel,modelPrefix,trainAnno.split('/')[-1],lr,numiter,NUM_WORKERS, IMS_PER_BATCH, BATCH_SIZE_PER_IMAGE, modelName) 
    
    with open(testAnno) as f:
      j = json.load(f)
    
    numClasses = len(j['categories'])
    # configure 
    cfg = get_cfg()
    cfg.merge_from_file(model_zoo.get_config_file("{}/{}.yaml".format(rootmodel, modelName)))
    cfg.MODEL.WEIGHTS = weights
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.7   # set the testing threshold 
    #cfg.MODEL.TENSOR_MASK.SCORE_THRESH_TEST = 0.8
    cfg.DATASETS.TEST = (name, )
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = numClasses
    cfg.MODEL.DEVICE='cpu' # run on cpu
    predictor = DefaultPredictor(cfg)

    for d in range(len(testdict)):
        l = testdict[d]['file_name'].split('/')
        l[0] = imgDataDir
        testdict[d]['file_name']='/'.join(l)
    
    MetadataCatalog.get(name).set(thing_classes=[j['categories'][i]['name'] for i in range(len(j['categories']))])
    #print(MetadataCatalog.get(name).thing_classes)
    if saveimages:
        plotfigs3(testdict,name,predictor,cfg,cat,fontSize=fontSize,alpha=alpha,threshold=threshold)
    else:
        plotfigs2(testdict,name,predictor,cfg,fontSize=fontSize,alpha=alpha,threshold=threshold)


def predictone(cat,trainjson,testjson,name,testdict,lr=0.00025,imgDataDir = '../', imgPath='',fontSize=5,alpha=0.1,threshold=0.7):
    #trainAnno='labels_json/use_train.json'
    #testAnno='labels_json/use_test_SalinaCruz_test.json'
    #testAnno='labels_json/use_test_Juchitan_test.json'
    trainAnno=trainjson#'labels_json/{}'.format(trainjson)
    
    
    try: register_coco_instances("trainset", {}, testjson, imgDataDir)
    except: pass

    NUM_WORKERS = 2
    IMS_PER_BATCH = 2
    BATCH_SIZE_PER_IMAGE = 128
    rootmodel = 'COCO-InstanceSegmentation'
    modelName='mask_rcnn_R_101_DC5_3x'
    modelPrefix=cat
    
    
    #lr=0.00025
    numiter=120000
    
    weights = "model/{}/{}_{}_{}_{}_{}_{}_{}_{}.pth".format(rootmodel,modelPrefix,trainAnno.split('/')[-1],lr,numiter,NUM_WORKERS, IMS_PER_BATCH, BATCH_SIZE_PER_IMAGE, modelName) 
    
    with open(testjson) as f:
      j = json.load(f)
    
    numClasses = len(j['categories'])
    # configure 
    cfg = get_cfg()
    cfg.merge_from_file(model_zoo.get_config_file("{}/{}.yaml".format(rootmodel, modelName)))
    cfg.MODEL.WEIGHTS = weights
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = threshold  # set the testing threshold 
    #cfg.MODEL.TENSOR_MASK.SCORE_THRESH_TEST = 0.8
    cfg.DATASETS.TEST = (name, )
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = numClasses
    cfg.MODEL.DEVICE='cpu' # run on cpu
    predictor = DefaultPredictor(cfg)


    
    MetadataCatalog.get(name).set(thing_classes=[j['categories'][i]['name'] for i in range(len(j['categories']))])



    if not os.path.exists('images/{}'.format(cat)):
        os.makedirs('images/{}'.format(cat))
    imgDir='images/{}/{}'



    if not os.path.exists(imgPath): 
        print(imgPath)
        print('Image doesn\'t exist.')
        return
    
    img = cv2.imread(imgPath)

    
    outputs = predictor(img)
    v = Visualizer(img[:, :, ::-1], MetadataCatalog.get(cfg.DATASETS.TEST[0]), scale=1)
    v._default_font_size = max(np.sqrt(v.output.height * v.output.width) // 90 * fontSize, 10 // 1.2 * fontSize)
    out = v.draw_instance_predictions(outputs["instances"].to("cpu"), alpha=alpha)

    w2h = v.output.width / v.output.height
    plt.figure(figsize=(10, 10/w2h))
    plt.imshow(out.get_image()[:, :, ::-1])
    #plt.title('Prediction')
    plt.axis("off")
    plt.rcParams['axes.xmargin'] = 0

    l = imgPath.split('/')
    
    plt.savefig('-'.join(l[-4:]),bbox_inches='tight')


import argparse

import torch
import torchvision.transforms as transforms
from torch.utils.data import DataLoader
import os
import csv

from models.resnet_applied import resnet50
from utils.Datasets import Foundation_Type_Testset

parser = argparse.ArgumentParser(description='Detect Foundation Type')

parser.add_argument('--image-path',help='Path to one image or a folder containing images.',required=True)
parser.add_argument('--checkpoint', default=None,type=str,
                    help='Path to checkpoint. Defaults to best pretrained version.')
parser.add_argument('--only-cpu', action='store_true', help='Use CPU only, disregard GPU.')
parser.add_argument('--mask-buildings', action='store_true')
parser.add_argument('--load-masks', action='store_true')
# This is no longer used 
#parser.add_argument('--model',help='Pretrained model, options ["foundation_v0.1"]', type=str)

os.makedirs('checkpoints',exist_ok=True)
weight_file_path = os.path.join('checkpoints','best_masked.pkl')

if not os.path.isfile(weight_file_path):
    print('Loading remote model file to the weights folder..')
    torch.hub.download_url_to_file('https://zenodo.org/record/4145934/files/best_masked.pkl', weight_file_path)
    
model_name='ade20k-resnet50dilated-ppm_deepsup'
model_path=os.path.join('csail_segmentation_tool','csail_seg',model_name)

encoder='{}/encoder_epoch_20.pth'.format(model_name)
decoder='{}/decoder_epoch_20.pth'.format(model_name)

os.makedirs(model_path,exist_ok=True)

if not os.path.isfile(os.path.join(model_path,'encoder_epoch_20.pth')):
    print('Loading remote model file to the weights folder..')
    torch.hub.download_url_to_file('http://sceneparsing.csail.mit.edu/model/pytorch/{}'.format(encoder),os.path.join(model_path,'encoder_epoch_20.pth'))

if not os.path.isfile(os.path.join(model_path,'decoder_epoch_20.pth')):
    print('Loading remote model file to the weights folder..')
    torch.hub.download_url_to_file('http://sceneparsing.csail.mit.edu/model/pytorch/{}'.format(decoder),os.path.join(model_path,'decoder_epoch_20.pth'))

args = parser.parse_args()

if args.only_cpu:
    device='cpu'
else:
    device = 'cuda' if torch.cuda.is_available() else 'cpu'

def main():
    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])

    test_transforms = [transforms.Resize((224, 224)),
                      transforms.ToTensor(),
                      normalize]

    if args.mask_buildings and not args.load_masks:
        from csail_segmentation_tool.csail_segmentation import MaskBuilding
        test_transforms.insert(0,
                              MaskBuilding(device))

    test_transforms = transforms.Compose(test_transforms)
    dataset = Foundation_Type_Testset(args.image_path, transform=test_transforms, mask_buildings=args.mask_buildings,
                                      load_masks = args.load_masks)

    test_loader = DataLoader(dataset, batch_size=1, shuffle=False, num_workers=0)
    model = resnet50(low_dim=1)
    
    if args.checkpoint is not None:
        model_file = args.checkpoint
    else:
        weight_file_path = os.path.join('checkpoints','best_model_weights.pth')
        if not os.path.isfile(weight_file_path):
            print('Loading remote model file to the weights folder..')
            torch.hub.download_url_to_file('https://zenodo.org/record/4145934/files/best_masked.pkl', weight_file_path)
        model_file = weight_file_path

    if not torch.cuda.is_available():
        state_dict = torch.load(model_file, map_location=torch.device('cpu'))
    else:
        state_dict = torch.load(model_file)


    missing, unexpected = model.load_state_dict(state_dict,
                                                strict=False)
    if any(['module' in name for name in unexpected]):
        # Remapping to remove effects of DataParallel
        from collections import OrderedDict
        new_state_dict = OrderedDict()
        for k, v in state_dict.items():
            name = k[7:]  # remove 'module.' of dataparallel
            new_state_dict[name] = v

        model.load_state_dict(new_state_dict,
                              strict=False)
    else:
        if len(missing) or len(unexpected):
            print('Missing or unexpected keys: {},{}'.format(missing, unexpected))
            print('This should not happen. Check if checkpoint is correct')

    model.eval()
    model = model.to(device)

    predictions = []

    with torch.no_grad():
        for i, (images, filename) in enumerate(test_loader):
            print('Classifying {}'.format(filename))
            images = images.to(device)

            prediction = model(images.float())
            predictions.append({'filename':filename[0],'prediction':int(torch.sigmoid(prediction).cpu().numpy() > 0.5)})

    with open('{}_prediction_results.csv'.format(os.path.basename(os.path.normpath(args.image_path))), 'w', newline='') as myfile:
        for prediction in predictions:
            wr = csv.writer(myfile, quoting=csv.QUOTE_ALL)
            wr.writerow([str(prediction['filename']),str(prediction['prediction'])])

    print ('Classification finished. Results written to {}'.format('{}_prediction_results.csv'.format(os.path.basename(os.path.normpath(args.image_path)))))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
/*------------------------------------------------------*
|                         BRAILS                        |
|                                                       |
| Author: Charles Wang,  UC Berkeley, c_w@berkeley.edu  |
|                                                       |
| Date:    10/15/2020                                   |
*------------------------------------------------------*/
"""

from brails.modules.ModelZoo import zoo
from brails.modules.GenericImageClassifier.GenericImageClassifier import *
import wget 
import os

class RoofClassifier(ImageClassifier):
    """ Roof Image Classifier. """


    def __init__(self, modelName=None, classNames=None, resultFile='roofType_preds.csv', workDir='tmp', printRes=True):
        '''
        modelFile: path to the model
        classNames: a list of classnames
        '''

        if not os.path.exists(workDir): os.makedirs(workDir)

        fileURL = zoo['roofType']['fileURL']
        
        if not classNames:
            classNames = zoo['roofType']['classNames']

        if not modelName:
            modelName = 'roof_classifier_v0.1'
            print('A default roof type model will be used: {}.'.format(modelName))

        modelFile = os.path.join(workDir,'{}.h5'.format(modelName))


        if not os.path.exists(modelFile): # download
            print('Downloading the model ...')
            downloadedModelFile = wget.download(fileURL, out=modelFile)

        ImageClassifier.__init__(self, modelName=modelName, classNames=classNames, resultFile=resultFile, printRes=printRes)



if __name__ == '__main__':
    main()



'''
Potential errors:
https://stackoverflow.com/questions/52805115/certificate-verify-failed-unable-to-get-local-issuer-certificate
'''




import argparse
import os
import sys

import torch
from PIL import Image


from csail_segmentation_tool.csail_segmentation import MaskBuilding
import numpy as np

parser = argparse.ArgumentParser(description='Detect Foundataion Type')

parser.add_argument('--image-path',help='Path to one image or a folder containing images.',required=True)
parser.add_argument('--only-cpu', action='store_true')
args = parser.parse_args()

if args.only_cpu:
    device='cpu'
else:
    device = 'cuda' if torch.cuda.is_available() else 'cpu'

def main(folder):
    if not args.only_cpu:
        print ('Will try to use available GPU.')
    mask_module = MaskBuilding(device)

    for root, _, fnames in sorted(os.walk(folder, followlinks=True)):
        for fname in sorted(fnames):

            if 'jpg' in fname or 'png' in fname:
                if 'mask' in fname:
                    print('Skipping {} because it has mask in filename. Presumably this is a mask already'.format(fname))
                    continue
                img_path = os.path.join(root, fname)

                _, file_extension = os.path.splitext(img_path)
                mask_filename = fname.replace(file_extension,'-mask.png')
                mask_path = os.path.join(root, mask_filename)
                if os.path.isfile(mask_path):
                    print('Skipping {} because mask for this file exists.'.format(fname))
                    continue

                try:
                    # Save the mask at the same name but add a '-mask suffix
                    with torch.no_grad():
                        mask = mask_module.get_mask(Image.open(img_path)).detach().cpu().squeeze().numpy()
                        mask_image = Image.fromarray(np.uint8(mask))
                        mask_image.save(mask_path)
                except RuntimeError as er:
                    if 'CUDA' in str(er):
                        print ('Error using GPU. If GPU exists maybe there is not enough GPU RAM. Try --only-cpu flag!')
                        sys.exit(1)
                    else:
                        print('Failed with unknown error: {}'.format(er))
                        sys.exit(1)


if __name__ =='__main__':
    main(args.image_path)



import argparse
import datetime
import os
import matplotlib as mpl
from attention_utils.utils import evaluate
import numpy as np

mpl.use('Agg')
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from torchvision import transforms
from attention_utils.radam import RAdam

from models.resnet_applied import resnet50

from torch.optim.lr_scheduler import StepLR

from utils.Datasets import Foundation_Type_Binary

torch.manual_seed(1337)
if torch.cuda.is_available():
    torch.cuda.manual_seed(1337)
np.random.seed(1337)

parser = argparse.ArgumentParser(description='Train Residual Attention')
parser.add_argument('--workers', type=int, default=8)
parser.add_argument('--epochs', default=500, type=int,
                    help='Train for n epochs')
parser.add_argument('--start-epoch', default=0, type=int,
                    help='Start epoch')
parser.add_argument('--lr', default=0.001, type=float,
                    help='Learning rate')
parser.add_argument('--batch-size', default=4, type=int,
                    help='Size of mini batch')
parser.add_argument('--train-data', type=str,
                    help='Training csv file with image names and labels')
parser.add_argument('--val-data', type=str,
                    help='Validation  csv file with image names and labels')
parser.add_argument('--test-data', type=str,
                    help='Test csv file with image names and labels')
parser.add_argument('--eval', action='store_true',
                    help='Only do evaluation')
parser.add_argument('--exp-name', default='noname', type=str, help='Name of this experiment to add to path name of log')
parser.add_argument('--checkpoint', type=str,
                    help='Path to checkpoint for evaluation or pretraining')
parser.add_argument('--mask-buildings', action='store_true')
parser.add_argument('--freeze-layers', action='store_true')
parser.add_argument('--pretrained', action='store_true', help='Use pretrained model.')

args = parser.parse_args()
if args.eval:
    args.pretrained = True
device = 'cuda' if torch.cuda.is_available() else 'cpu'

currentTime = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")

log_dir = os.path.join('runs', 'run_{}'.format(args.exp_name) + currentTime)
summary_writer = SummaryWriter(log_dir)

lr = args.lr  # 0.1
summary_writer.add_text('Parameters', str(args))
if args.start_epoch > 0:
    args.epochs = args.start_epoch + args.epochs

def main():
    best_checkpoint = os.path.join(log_dir, 'checkpoints/best_self_trained.pkl')

    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])

    train_transforms = [
        transforms.RandomResizedCrop(224, scale=(0.7, 1.)),
        transforms.RandomGrayscale(p=0.5),
        transforms.ColorJitter(0.5, 0.5, 0.5, 0.5),
        transforms.RandomHorizontalFlip(),
        transforms.RandomVerticalFlip(),
        transforms.ToTensor(),
        normalize]

    train_transforms = transforms.Compose(train_transforms)

    val_transforms = [transforms.Resize((224, 224)),
        transforms.ToTensor(),
        normalize]

    val_transforms = transforms.Compose(val_transforms)

    if not args.eval:
        
        train_dataset = Foundation_Type_Binary(args.train_data,
                                               transform=train_transforms,
                                               mask_buildings=args.mask_buildings, load_masks=True)

        val_dataset = Foundation_Type_Binary(args.val_data,
                                             transform=val_transforms,
                                             mask_buildings=args.mask_buildings, load_masks=True)
        
        train_weights = np.array(train_dataset.train_weights)
        train_sampler = torch.utils.data.WeightedRandomSampler(train_weights, len(train_weights),
                                                                   replacement=True)
    

        train_loader = DataLoader(train_dataset, batch_size=args.batch_size, num_workers=args.workers,
                                  drop_last=False, sampler=train_sampler)
        val_loader = DataLoader(val_dataset, batch_size=args.batch_size, shuffle=False, num_workers=args.workers,
                                drop_last=False)

    else:
        test_dataset = Foundation_Type_Binary(args.test_data,
                                             transform=val_transforms,
                                             mask_buildings=args.mask_buildings, load_masks=True)
        test_loader = DataLoader(test_dataset, batch_size=args.batch_size, shuffle=False, num_workers=args.workers,
                                drop_last=False)

    model = resnet50(low_dim=1)

    # Freeze all layers apart from the final layer
    if args.freeze_layers:
        ct = 0
        for child in model.children():
            ct += 1
            if ct < 10:
                print('Freezing {}'.format(child))
                for param in child.parameters():
                    param.requires_grad = False

    summary_writer.add_text('Architecture', model.__class__.__name__)
    summary_writer.add_text('Train Transforms', str(train_transforms))
    summary_writer.add_text('Val Transforms', str(val_transforms))

    criterion = nn.BCEWithLogitsLoss()

    optimizer = RAdam(model.parameters(), lr=lr)
    scheduler = StepLR(optimizer, step_size=1, gamma=0.9)

    summary_writer.add_text('Criterion', str(criterion))
    summary_writer.add_text('Optimizer', str(optimizer))

    is_train = not args.eval

    best_perf = 1e6

    model = nn.DataParallel(model).to(device)
    if args.pretrained:
        try:
            state_dict = torch.load(args.checkpoint)['state_dict']
        except KeyError: # Format of NPID checkpoints, and checkpoints created with train.py differ
            state_dict = torch.load(args.checkpoint)

        missing, unexpected = model.load_state_dict(state_dict,
                                                    strict=False)
        if len(missing) or len(unexpected):
            print('Missing or unexpected keys: {},{}'.format(missing, unexpected))

    if is_train is True:
        for epoch in range(args.start_epoch, args.epochs):
            model.train()

            print('Training epoch: {}'.format(epoch))
            y_train_pred, y_train_gt, avg_train_loss = parse(model, train_loader, criterion, optimizer, 'train', epoch)
            scheduler.step()
            evaluate(summary_writer, 'Train', y_train_gt, y_train_pred, avg_train_loss,
                     train_loader.dataset.classes, epoch)

            print('Validation epoch: {}'.format(epoch))
            with torch.no_grad():
                y_val_pred, y_val_gt, avg_val_loss = parse(model, val_loader, criterion, None, 'val', epoch)

            current_perf = evaluate(summary_writer, 'Val',
                                    y_val_gt, y_val_pred, avg_val_loss,
                                    train_loader.dataset.classes, epoch)

            if current_perf > best_perf:
                best_perf = current_perf
                print('current best performance measure,', best_perf)
                torch.save(model.state_dict(), best_checkpoint)

            # Save regular checkpoint every epoch
            latest_model_path = os.path.join(log_dir, 'checkpoint_epoch_{}.pkl'.format(epoch))
            torch.save(model.state_dict(), latest_model_path)
        print('best performance measure mse' + str(best_perf))

    else:
        print('Only test mode:')
        with torch.no_grad():
            y_val_pred, y_val_gt, avg_val_loss = parse(model, test_loader, criterion, None, 'test', 0)

        current_perf = evaluate(summary_writer, 'Test',
                                y_val_gt, y_val_pred, avg_val_loss,
                                test_loader.dataset.classes, 0)

        print('F1: {}'.format(current_perf[2]))
        print('Precision: {}'.format(current_perf[0]))
        print('Recall: {}'.format(current_perf[1]))
        exit()


def parse(model, data_loader, criterion, optimizer, parse_type, epoch):
    train = False
    if parse_type == 'train':
        model.train()
        train = True
    elif parse_type == 'val' or parse_type == 'test':
        model.eval()
    else:
        raise ValueError('Wrong parse type. Options are train, val or test')

    predictions = []
    ground_truths = []

    avg_loss = 0

    for i, (images, labels, _) in enumerate(data_loader):
        images = images.to(device)
        labels = labels.to(device)

        if train:
            optimizer.zero_grad()

        prediction = model(images.float())
        loss = criterion(prediction.squeeze(), labels.squeeze())

        if train:
            loss.backward()
            optimizer.step()

        avg_loss += loss.detach().cpu().numpy()

        predictions.extend(torch.sigmoid(prediction).detach().cpu().tolist())  # Should do the detach internally
        ground_truths.extend(labels.detach().cpu().tolist())

        if (i + 1) % 100 == 0:
            print("Epoch [%d/%d], Iter [%d/%d] Loss: %.4f" % (
                epoch + 1, args.epochs, i + 1, len(data_loader), loss.data.item()))

    return predictions, ground_truths, avg_loss / len(data_loader)


if __name__ == '__main__':
    main()



# System libs
import os
import argparse
from distutils.version import LooseVersion
from multiprocessing import Queue, Process
# Numerical libs
import numpy as np
import math
import torch
import torch.nn as nn
from scipy.io import loadmat
# Our libs
from config import cfg
from dataset import ValDataset
from models import ModelBuilder, SegmentationModule
from utils import AverageMeter, colorEncode, accuracy, intersectionAndUnion, parse_devices, setup_logger
from lib.nn import user_scattered_collate, async_copy_to
from lib.utils import as_numpy
from PIL import Image
from tqdm import tqdm

colors = loadmat('data/color150.mat')['colors']


def visualize_result(data, pred, dir_result):
    (img, seg, info) = data

    # segmentation
    seg_color = colorEncode(seg, colors)

    # prediction
    pred_color = colorEncode(pred, colors)

    # aggregate images and save
    im_vis = np.concatenate((img, seg_color, pred_color),
                            axis=1).astype(np.uint8)

    img_name = info.split('/')[-1]
    Image.fromarray(im_vis).save(os.path.join(dir_result, img_name.replace('.jpg', '.png')))


def evaluate(segmentation_module, loader, cfg, gpu_id, result_queue):
    segmentation_module.eval()

    for batch_data in loader:
        # process data
        batch_data = batch_data[0]
        seg_label = as_numpy(batch_data['seg_label'][0])
        img_resized_list = batch_data['img_data']

        with torch.no_grad():
            segSize = (seg_label.shape[0], seg_label.shape[1])
            scores = torch.zeros(1, cfg.DATASET.num_class, segSize[0], segSize[1])
            scores = async_copy_to(scores, gpu_id)

            for img in img_resized_list:
                feed_dict = batch_data.copy()
                feed_dict['img_data'] = img
                del feed_dict['img_ori']
                del feed_dict['info']
                feed_dict = async_copy_to(feed_dict, gpu_id)

                # forward pass
                scores_tmp = segmentation_module(feed_dict, segSize=segSize)
                scores = scores + scores_tmp / len(cfg.DATASET.imgSizes)

            _, pred = torch.max(scores, dim=1)
            pred = as_numpy(pred.squeeze(0).cpu())

        # calculate accuracy and SEND THEM TO MASTER
        acc, pix = accuracy(pred, seg_label)
        intersection, union = intersectionAndUnion(pred, seg_label, cfg.DATASET.num_class)
        result_queue.put_nowait((acc, pix, intersection, union))

        # visualization
        if cfg.VAL.visualize:
            visualize_result(
                (batch_data['img_ori'], seg_label, batch_data['info']),
                pred,
                os.path.join(cfg.DIR, 'result')
            )


def worker(cfg, gpu_id, start_idx, end_idx, result_queue):
    torch.cuda.set_device(gpu_id)

    # Dataset and Loader
    dataset_val = ValDataset(
        cfg.DATASET.root_dataset,
        cfg.DATASET.list_val,
        cfg.DATASET,
        start_idx=start_idx, end_idx=end_idx)
    loader_val = torch.utils.data.DataLoader(
        dataset_val,
        batch_size=cfg.VAL.batch_size,
        shuffle=False,
        collate_fn=user_scattered_collate,
        num_workers=2)

    # Network Builders
    net_encoder = ModelBuilder.build_encoder(
        arch=cfg.MODEL.arch_encoder.lower(),
        fc_dim=cfg.MODEL.fc_dim,
        weights=cfg.MODEL.weights_encoder)
    net_decoder = ModelBuilder.build_decoder(
        arch=cfg.MODEL.arch_decoder.lower(),
        fc_dim=cfg.MODEL.fc_dim,
        num_class=cfg.DATASET.num_class,
        weights=cfg.MODEL.weights_decoder,
        use_softmax=True)

    crit = nn.NLLLoss(ignore_index=-1)

    segmentation_module = SegmentationModule(net_encoder, net_decoder, crit)

    segmentation_module.cuda()

    # Main loop
    evaluate(segmentation_module, loader_val, cfg, gpu_id, result_queue)


def main(cfg, gpus):
    with open(cfg.DATASET.list_val, 'r') as f:
        lines = f.readlines()
        num_files = len(lines)

    num_files_per_gpu = math.ceil(num_files / len(gpus))

    pbar = tqdm(total=num_files)

    acc_meter = AverageMeter()
    intersection_meter = AverageMeter()
    union_meter = AverageMeter()

    result_queue = Queue(500)
    procs = []
    for idx, gpu_id in enumerate(gpus):
        start_idx = idx * num_files_per_gpu
        end_idx = min(start_idx + num_files_per_gpu, num_files)
        proc = Process(target=worker, args=(cfg, gpu_id, start_idx, end_idx, result_queue))
        print('gpu:{}, start_idx:{}, end_idx:{}'.format(gpu_id, start_idx, end_idx))
        proc.start()
        procs.append(proc)

    # master fetches results
    processed_counter = 0
    while processed_counter < num_files:
        if result_queue.empty():
            continue
        (acc, pix, intersection, union) = result_queue.get()
        acc_meter.update(acc, pix)
        intersection_meter.update(intersection)
        union_meter.update(union)
        processed_counter += 1
        pbar.update(1)

    for p in procs:
        p.join()

    # summary
    iou = intersection_meter.sum / (union_meter.sum + 1e-10)
    for i, _iou in enumerate(iou):
        print('class [{}], IoU: {:.4f}'.format(i, _iou))

    print('[Eval Summary]:')
    print('Mean IoU: {:.4f}, Accuracy: {:.2f}%'
          .format(iou.mean(), acc_meter.average()*100))

    print('Evaluation Done!')


if __name__ == '__main__':
    assert LooseVersion(torch.__version__) >= LooseVersion('0.4.0'), \
        'PyTorch>=0.4.0 is required'

    parser = argparse.ArgumentParser(
        description="PyTorch Semantic Segmentation Validation"
    )
    parser.add_argument(
        "--cfg",
        default="config/ade20k-resnet50dilated-ppm_deepsup.yaml",
        metavar="FILE",
        help="path to config file",
        type=str,
    )
    parser.add_argument(
        "--gpus",
        default="0-3",
        help="gpus to use, e.g. 0-3 or 0,1,2,3"
    )
    parser.add_argument(
        "opts",
        help="Modify config options using the command-line",
        default=None,
        nargs=argparse.REMAINDER,
    )
    args = parser.parse_args()

    cfg.merge_from_file(args.cfg)
    cfg.merge_from_list(args.opts)
    # cfg.freeze()

    logger = setup_logger(distributed_rank=0)   # TODO
    logger.info("Loaded configuration file {}".format(args.cfg))
    logger.info("Running with config:\n{}".format(cfg))

    # absolute paths of model weights
    cfg.MODEL.weights_encoder = os.path.join(
        cfg.DIR, 'encoder_' + cfg.VAL.checkpoint)
    cfg.MODEL.weights_decoder = os.path.join(
        cfg.DIR, 'decoder_' + cfg.VAL.checkpoint)
    assert os.path.exists(cfg.MODEL.weights_encoder) and \
        os.path.exists(cfg.MODEL.weights_decoder), "checkpoint does not exitst!"

    if not os.path.isdir(os.path.join(cfg.DIR, "result")):
        os.makedirs(os.path.join(cfg.DIR, "result"))

    # Parse gpu ids
    gpus = parse_devices(args.gpus)
    gpus = [x.replace('gpu', '') for x in gpus]
    gpus = [int(x) for x in gpus]

    main(cfg, gpus)



# Author: Zylo117

import math

import torch
from torch import nn

from lib.efficientdet.model import BiFPN, Regressor, Classifier, EfficientNet
from lib.efficientdet.utils import Anchors


class EfficientDetBackbone(nn.Module):
    def __init__(self, num_classes=80, compound_coef=0, load_weights=False, **kwargs):
        super(EfficientDetBackbone, self).__init__()
        self.compound_coef = compound_coef

        self.backbone_compound_coef = [0, 1, 2, 3, 4, 5, 6, 6]
        self.fpn_num_filters = [64, 88, 112, 160, 224, 288, 384, 384]
        self.fpn_cell_repeats = [3, 4, 5, 6, 7, 7, 8, 8]
        self.input_sizes = [512, 640, 768, 896, 1024, 1280, 1280, 1536]
        self.box_class_repeats = [3, 3, 3, 4, 4, 4, 5, 5]
        self.anchor_scale = [4., 4., 4., 4., 4., 4., 4., 5.]
        self.aspect_ratios = kwargs.get('ratios', [(1.0, 1.0), (1.4, 0.7), (0.7, 1.4)])
        self.num_scales = len(kwargs.get('scales', [2 ** 0, 2 ** (1.0 / 3.0), 2 ** (2.0 / 3.0)]))
        conv_channel_coef = {
            # the channels of P3/P4/P5.
            0: [40, 112, 320],
            1: [40, 112, 320],
            2: [48, 120, 352],
            3: [48, 136, 384],
            4: [56, 160, 448],
            5: [64, 176, 512],
            6: [72, 200, 576],
            7: [72, 200, 576],
        }

        num_anchors = len(self.aspect_ratios) * self.num_scales

        self.bifpn = nn.Sequential(
            *[BiFPN(self.fpn_num_filters[self.compound_coef],
                    conv_channel_coef[compound_coef],
                    True if _ == 0 else False,
                    attention=True if compound_coef < 6 else False)
              for _ in range(self.fpn_cell_repeats[compound_coef])])

        self.num_classes = num_classes
        self.regressor = Regressor(in_channels=self.fpn_num_filters[self.compound_coef], num_anchors=num_anchors,
                                   num_layers=self.box_class_repeats[self.compound_coef])
        self.classifier = Classifier(in_channels=self.fpn_num_filters[self.compound_coef], num_anchors=num_anchors,
                                     num_classes=num_classes,
                                     num_layers=self.box_class_repeats[self.compound_coef])

        self.anchors = Anchors(anchor_scale=self.anchor_scale[compound_coef], **kwargs)

        self.backbone_net = EfficientNet(self.backbone_compound_coef[compound_coef], load_weights)

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def forward(self, inputs):
        max_size = inputs.shape[-1]

        _, p3, p4, p5 = self.backbone_net(inputs)

        features = (p3, p4, p5)
        features = self.bifpn(features)

        regression = self.regressor(features)
        classification = self.classifier(features)
        anchors = self.anchors(inputs, inputs.dtype)

        return features, regression, classification, anchors

    def init_backbone(self, path):
        state_dict = torch.load(path)
        try:
            ret = self.load_state_dict(state_dict, strict=False)
            print(ret)
        except RuntimeError as e:
            print('Ignoring ' + str(e) + '"')



import sys
import os
try:
    from urllib import urlretrieve
except ImportError:
    from urllib.request import urlretrieve
import torch


def load_url(url, model_dir='./pretrained', map_location=None):
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)
    filename = url.split('/')[-1]
    cached_file = os.path.join(model_dir, filename)
    if not os.path.exists(cached_file):
        sys.stderr.write('Downloading: "{}" to {}\n'.format(url, cached_file))
        urlretrieve(url, cached_file)
    return torch.load(cached_file, map_location=map_location)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 The Regents of the University of California
#
# This file is part of BRAILS.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# BRAILS. If not, see <http://www.opensource.org/licenses/>.
#
# Author:
# Chaofeng Wang, c_w@berkeley.edu


from brails.modules.GenericImageClassifier.GenericImageClassifier import ImageClassifier
from brails.modules.SoftstoryClassifier.SoftstoryClassifier import SoftstoryClassifier as SoftstoryClassifier
from brails.modules.RoofTypeClassifier.RoofTypeClassifier import RoofClassifier as RoofClassifier
from brails.modules.OccupancyClassClassifier.OccupancyClassClassifier import OccupancyClassifier as OccupancyClassifier
from brails.modules.Foundation_Classification.FoundationClassifier import FoundationHeightClassifier
from brails.modules.Year_Built_Classifier.YearBuiltClassifier import YearBuiltClassifier





from lib.train_detector import Detector

import os
import cv2
import numpy as np
from shapely.geometry import Polygon
from shapely.ops import unary_union
from lib.infer_detector import Infer
import torch
import time
from tqdm import tqdm

class NFloorDetector():
    def __init__(self):
        self.system_dict = {}
        self.system_dict["train"] = {}
        self.system_dict["train"]["data"] = {}
        self.system_dict["train"]["model"] = {}        
        self.system_dict["infer"] = {}
        self.system_dict["infer"]["params"] = {}

        self.set_fixed_params()
    
    def set_fixed_params(self):        
        self.system_dict["train"]["data"]["trainSet"] = "train"
        self.system_dict["train"]["data"]["validSet"] = "valid"
        self.system_dict["train"]["data"]["classes"] = ["floor"]
        self.system_dict["train"]["model"]["valInterval"] = 1
        self.system_dict["train"]["model"]["saveInterval"] = 5
        self.system_dict["train"]["model"]["esMinDelta"] = 0.0
        self.system_dict["train"]["model"]["esPatience"] = 0
        
    def load_train_data(self, rootDir="datasets/", nWorkers=0, batchSize=2):
        self.system_dict["train"]["data"]["rootDir"] = rootDir
        self.system_dict["train"]["data"]["nWorkers"] = nWorkers
        self.system_dict["train"]["data"]["batchSize"] = batchSize        
        

    def train(self, compCoeff=3, topOnly=False, optim="adamw", lr=1e-4, numEpochs=25, nGPU=1):
        self.system_dict["train"]["model"]["compCoeff"] = compCoeff
        self.system_dict["train"]["model"]["topOnly"] = topOnly
        self.system_dict["train"]["model"]["optim"] = optim          
        self.system_dict["train"]["model"]["lr"] = lr
        self.system_dict["train"]["model"]["numEpochs"] = numEpochs
        self.system_dict["train"]["model"]["nGPU"] = nGPU        
        
        # Create the Object Detector Object
        gtf = Detector()

        gtf.set_train_dataset(self.system_dict["train"]["data"]["rootDir"],
                              "",
                              "",
                              self.system_dict["train"]["data"]["trainSet"],
                              classes_list=self.system_dict["train"]["data"]["classes"],
                              batch_size=self.system_dict["train"]["data"]["batchSize"],
                              num_workers=self.system_dict["train"]["data"]["nWorkers"])        

        gtf.set_val_dataset(self.system_dict["train"]["data"]["rootDir"],
                            "",
                            "",
                            self.system_dict["train"]["data"]["validSet"])
        
        # Define the Model Architecture
        coeff = self.system_dict["train"]["model"]["compCoeff"]
        modelArchitecture = f"efficientdet-d{coeff}.pth"
        
        gtf.set_model(model_name=modelArchitecture,
                      num_gpus=self.system_dict["train"]["model"]["nGPU"],
                      freeze_head=self.system_dict["train"]["model"]["topOnly"])
        
        # Set Model Hyperparameters    
        gtf.set_hyperparams(optimizer=self.system_dict["train"]["model"]["optim"], 
                            lr=self.system_dict["train"]["model"]["lr"],
                            es_min_delta=self.system_dict["train"]["model"]["esMinDelta"], 
                            es_patience=self.system_dict["train"]["model"]["esPatience"])
        
        # Train    
        gtf.train(num_epochs=self.system_dict["train"]["model"]["numEpochs"],
                  val_interval=self.system_dict["train"]["model"]["valInterval"],
                  save_interval=self.system_dict["train"]["model"]["saveInterval"])

    def infer(self, imPath="datasets/test/", modelPath="models/efficientdet-d4_trained.pth", gpuEnabled=True, outFile="nFloorPredict.csv"):
        self.system_dict["infer"]["imPath"] = imPath
        self.system_dict["infer"]["modelPath"] = modelPath
        self.system_dict["infer"]["gpuEnabled"] = gpuEnabled
        self.system_dict["infer"]["outFile"] = outFile
        
        def install_default_model(model_path):
            if model_path == "models/efficientdet-d4_trained.pth":
                os.makedirs('models',exist_ok=True)
                model_path = os.path.join('models','efficientdet-d4_trained.pth')
        
                if not os.path.isfile(model_path):
                    print('Loading default model file to the models folder...')
                    torch.hub.download_url_to_file('https://zenodo.org/record/4421613/files/efficientdet-d4_trained.pth',
                                                   model_path, progress=False)
                    print('Default model loaded.')
            else:
                print(f'Inferences will be performed using the custom model at {model_path}.')
        
        def create_polygon(bb):
            polygon = Polygon([(bb[0], bb[1]), (bb[0], bb[3]),
                               (bb[2], bb[3]), (bb[2], bb[1])])
            return polygon
        
        def intersect_polygons(poly1, poly2):
            if poly1.intersects(poly2):
                polyArea = poly1.intersection(poly2).area
                if poly1.area!=0 and poly2.area!=0:
                    overlapRatio = polyArea/poly1.area*100
                else: 
                    overlapRatio = 0
            else:
                overlapRatio = 0
            return overlapRatio
        
        def check_threshold_level(boxesPoly):
            thresholdChange = False
            thresholdIncrease = False
            if not boxesPoly:
                thresholdChange = True
                thresholdIncrease = False
            else:
                falseDetect = np.zeros(len(boxesPoly))
                for k in range(len(boxesPoly)):
                    overlapRatio = np.array([intersect_polygons(p, boxesPoly[k]) for p in boxesPoly],dtype=float)
                    falseDetect[k] = len([idx for idx, val in enumerate(overlapRatio) if val > 75][1:])
                thresholdChange = any(falseDetect>2)
                if thresholdChange: thresholdIncrease = True
            return thresholdChange, thresholdIncrease
        
        def compute_derivative(centBoxes):
            nBoxes = centBoxes.shape[0] 
            dyOverdx = np.zeros((nBoxes,nBoxes)) + 10
            for k in range(nBoxes):
                for m in range(nBoxes):
                    dx = abs(centBoxes[k,0]-centBoxes[m,0])
                    dy = abs(centBoxes[k,1]-centBoxes[m,1])
                    if k!=m:
                        dyOverdx[k,m] = dy/dx
            return dyOverdx
        
        install_default_model(self.system_dict["infer"]["modelPath"])
        
        # Start Program Timer
        startTime = time.time()
        
        # Get the Image List
        imgList = os.listdir(self.system_dict["infer"]["imPath"])
        nImages = len(imgList)
        
        # Initiate a CSV File to Write Program Output If More Than One Image in imageDir
        if nImages!=1:
            csvFile = open(self.system_dict["infer"]["outFile"],'w+')
            csvFile.write("Image, nFloors\n")
            
        # Create and Define the Inference Model
        classes = ["floor"]
        
        print("Performing inferences on images...")
        gtfInfer = Infer()
        gtfInfer.load_model(self.system_dict["infer"]["modelPath"], classes, use_gpu=self.system_dict["infer"]["gpuEnabled"])
        for imgno in tqdm(range(6519,nImages)):
            # Perform Iterative Inference
            imgPath = os.path.join(self.system_dict["infer"]["imPath"],imgList[imgno])
            img = cv2.imread(imgPath)
            img = cv2.resize(img,(640,640))
            cv2.imwrite("input.jpg",img)
            _, _, boxes = gtfInfer.predict("input.jpg",threshold=0.2)
            boxesPoly = [create_polygon(bb) for bb in boxes]
            
            multiplier = 1
            while check_threshold_level(boxesPoly)[0]:
                if check_threshold_level(boxesPoly)[1]:
                    confThreshold = 0.2 + multiplier*0.1
                    if confThreshold>1: break
                else:
                    confThreshold = 0.2 - multiplier*0.02
                    if confThreshold==0: break
                _, _, boxes = gtfInfer.predict("input.jpg",threshold=confThreshold)                    
                multiplier +=1        
                boxesPoly = [create_polygon(bb) for bb in boxes]
            
            # Postprocessing    
            boxesPoly = [create_polygon(bb) for bb in boxes]
            
            nestedBoxes = np.zeros((10*len(boxes)),dtype=int)
            counter = 0
            for bbPoly in boxesPoly:    
                overlapRatio = np.array([intersect_polygons(p, bbPoly) for p in boxesPoly],dtype=float)
                ind = [idx for idx, val in enumerate(overlapRatio) if val > 75][1:]
                nestedBoxes[counter:counter+len(ind)] = ind
                counter += len(ind)
            nestedBoxes  = np.unique(nestedBoxes[:counter])
            
            counter = 0
            for x in nestedBoxes:
                del boxes[x-counter] 
                counter += 1
            
            nBoxes = len(boxes)
            
            boxesPoly = []
            boxesExtendedPoly = []
            centBoxes = np.zeros((nBoxes,2))
            for k in range(nBoxes):    
                bb = boxes[k]  
                tempPoly = create_polygon(bb)
                boxesPoly.append(tempPoly)
                x,y = tempPoly.centroid.xy
                centBoxes[k,:] = np.array([x[0],y[0]])
                boxesExtendedPoly.append(create_polygon([0.9*bb[0],0,1.1*bb[2],len(img)-1]))
            
            stackedInd = []
            for bb in boxesExtendedPoly:    
                overlapRatio = np.array([intersect_polygons(p, bb) for p in boxesExtendedPoly],dtype=float)
                stackedInd.append([idx for idx, val in enumerate(overlapRatio) if val > 10])
                
            uniqueStacks0 = [list(x) for x in set(tuple(x) for x in stackedInd)]
            
            dyOverdx = compute_derivative(centBoxes)
            stacks = np.where(dyOverdx>1.3)
            
            counter = 0
            uniqueStacks = [[] for i in range(nBoxes)]
            for k in range(nBoxes):
                while counter<len(stacks[0]) and k==stacks[0][counter]:
                    uniqueStacks[k].append(stacks[1][counter])
                    counter +=1
            
            uniqueStacks = [list(x) for x in set(tuple(x) for x in uniqueStacks)]
            
            if len(uniqueStacks0)==1 or len(uniqueStacks)==1:
                nFloors = len(uniqueStacks0[0])
            else:
                lBound = len(img)/5
                uBound = 4*len(img)/5
                middlePoly = Polygon([(lBound,0),(lBound,len(img)),(uBound,len(img)),(uBound,0)])
                overlapRatio = np.empty(len(uniqueStacks))
                for k in range(len(uniqueStacks)):
                    poly = unary_union([boxesPoly[x] for x in uniqueStacks[k]])
                    overlapRatio[k] = (intersect_polygons(poly, middlePoly))
            
                
                indKeep = np.argsort(-overlapRatio)[0:2]
                stack4Address = []
                for k in range(2):
                    if overlapRatio[indKeep[k]]>10: stack4Address.append(indKeep[k])
                if len(stack4Address)!=0:
                    nFloors = max([len(uniqueStacks[x]) for x in stack4Address])
                else:
                    nFloors = len(uniqueStacks[0])
                    
               
            if nImages!=1: 
                    csvFile.write(os.path.splitext(imgList[imgno])[0] + f", {nFloors}\n")
            else:
                print(f"Number of floors:{nFloors}\n")        
        
        csvFile.close()
        # End Program Timer and Display Execution Time
        endTime = time.time()
        hours, rem = divmod(endTime-startTime, 3600)
        minutes, seconds = divmod(rem, 60)
        print("\nTotal execution time: {:0>2}:{:0>2}:{:05.2f}".format(int(hours),int(minutes),seconds))
        
        # Cleanup the Root Folder
        if os.path.isfile("input.jpg"):
            os.remove("input.jpg")
        if os.path.isfile("output.jpg"):
            os.remove("output.jpg")
            


'''Train CIFAR10 with PyTorch.'''
from __future__ import print_function

import sys
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import torch.backends.cudnn as cudnn

import torchvision
import torchvision.transforms as transforms
import lib.custom_transforms as custom_transforms

import os
import argparse
import time

import models
import datasets
import math

from lib.NCEAverage import NCEAverage
from lib.LinearAverage import LinearAverage
from lib.NCECriterion import NCECriterion
from lib.utils import AverageMeter
from test import NN, kNN

parser = argparse.ArgumentParser(description='PyTorch CIFAR10 Training')
parser.add_argument('--lr', default=0.03, type=float, help='learning rate')
parser.add_argument('--resume', '-r', default='', type=str, help='resume from checkpoint')
parser.add_argument('--test-only', action='store_true', help='test only')
parser.add_argument('--low-dim', default=128, type=int,
                    metavar='D', help='feature dimension')
parser.add_argument('--nce-k', default=4096, type=int,
                    metavar='K', help='number of negative samples for NCE')
parser.add_argument('--nce-t', default=0.1, type=float,
                    metavar='T', help='temperature parameter for softmax')
parser.add_argument('--nce-m', default=0.5, type=float,
                    metavar='M', help='momentum for non-parametric updates')

args = parser.parse_args()

device = 'cuda' if torch.cuda.is_available() else 'cpu'
best_acc = 0  # best test accuracy
start_epoch = 0  # start from epoch 0 or last checkpoint epoch

# Data
print('==> Preparing data..')
transform_train = transforms.Compose([
    transforms.RandomResizedCrop(size=32, scale=(0.2,1.)),
    transforms.ColorJitter(0.4, 0.4, 0.4, 0.4),
    transforms.RandomGrayscale(p=0.2),
    #transforms.RandomHorizontalFlip(),
    transforms.ToTensor(),
    transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
])

transform_test = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
])

trainset = datasets.CIFAR10Instance(root='./data', train=True, download=True, transform=transform_train)
trainloader = torch.utils.data.DataLoader(trainset, batch_size=128, shuffle=True, num_workers=2)

testset = datasets.CIFAR10Instance(root='./data', train=False, download=True, transform=transform_test)
testloader = torch.utils.data.DataLoader(testset, batch_size=100, shuffle=False, num_workers=2)

classes = ('plane', 'car', 'bird', 'cat', 'deer', 'dog', 'frog', 'horse', 'ship', 'truck')
ndata = trainset.__len__()

print('==> Building model..')
net = models.__dict__['ResNet18'](low_dim=args.low_dim)
# define leminiscate
if args.nce_k > 0:
    lemniscate = NCEAverage(args.low_dim, ndata, args.nce_k, args.nce_t, args.nce_m)
else:
    lemniscate = LinearAverage(args.low_dim, ndata, args.nce_t, args.nce_m)

if device == 'cuda':
    net = torch.nn.DataParallel(net, device_ids=range(torch.cuda.device_count()))
    cudnn.benchmark = True

# Model
if args.test_only or len(args.resume)>0:
    # Load checkpoint.
    print('==> Resuming from checkpoint..')
    assert os.path.isdir('checkpoint'), 'Error: no checkpoint directory found!'
    checkpoint = torch.load('./checkpoint/'+args.resume)
    net.load_state_dict(checkpoint['net'])
    lemniscate = checkpoint['lemniscate']
    best_acc = checkpoint['acc']
    start_epoch = checkpoint['epoch']
    
# define loss function
if hasattr(lemniscate, 'K'):
    criterion = NCECriterion(ndata)
else:
    criterion = nn.CrossEntropyLoss()

net.to(device)
lemniscate.to(device)
criterion.to(device)

if args.test_only:
    acc = kNN(0, net, lemniscate, trainloader, testloader, 200, args.nce_t, 1)
    sys.exit(0)

optimizer = optim.SGD(net.parameters(), lr=args.lr, momentum=0.9, weight_decay=5e-4)

def adjust_learning_rate(optimizer, epoch):
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    lr = args.lr
    if epoch >= 80:
        lr = args.lr * (0.1 ** ((epoch-80) // 40))
    print(lr)
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr

# Training
def train(epoch):
    print('\nEpoch: %d' % epoch)
    adjust_learning_rate(optimizer, epoch)
    train_loss = AverageMeter()
    data_time = AverageMeter()
    batch_time = AverageMeter()
    correct = 0
    total = 0

    # switch to train mode
    net.train()

    end = time.time()
    for batch_idx, (inputs, targets, indexes) in enumerate(trainloader):
        data_time.update(time.time() - end)
        inputs, targets, indexes = inputs.to(device), targets.to(device), indexes.to(device)
        optimizer.zero_grad()

        features = net(inputs)
        outputs = lemniscate(features, indexes)
        loss = criterion(outputs, indexes)

        loss.backward()
        optimizer.step()

        train_loss.update(loss.item(), inputs.size(0))

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        print('Epoch: [{}][{}/{}]'
              'Time: {batch_time.val:.3f} ({batch_time.avg:.3f}) '
              'Data: {data_time.val:.3f} ({data_time.avg:.3f}) '
              'Loss: {train_loss.val:.4f} ({train_loss.avg:.4f})'.format(
              epoch, batch_idx, len(trainloader), batch_time=batch_time, data_time=data_time, train_loss=train_loss))

for epoch in range(start_epoch, start_epoch+200):
    train(epoch)
    acc = kNN(epoch, net, lemniscate, trainloader, testloader, 200, args.nce_t, 0)

    if acc > best_acc:
        print('Saving..')
        state = {
            'net': net.state_dict(),
            'lemniscate': lemniscate,
            'acc': acc,
            'epoch': epoch,
        }
        if not os.path.isdir('checkpoint'):
            os.mkdir('checkpoint')
        torch.save(state, './checkpoint/ckpt.t7')
        best_acc = acc

    print('best accuracy: {:.2f}'.format(best_acc*100))

acc = kNN(0, net, lemniscate, trainloader, testloader, 200, args.nce_t, 1)
print('last accuracy: {:.2f}'.format(acc*100))
